/*
** cardinfo.c
**
** Read the EEPROM on the card and update the members of ADMXRC::info 
*/

#include "device.h"
#include "eeprom.h"

/*
** Read EEPROM info on an ADM-XRC/ADM-XRC-P/ADM-XRC-IIL.
** Also find the base and limits of the PLX spaces S0, S1 on
** the local bus.
**
** There are some differences between the ADM-XRC/ADM-XRC-P
** and the ADM-XRC-IIL in the encoding of the EEPROM info,
** but there is sufficient commonality to handle both boards
** in this routine.
*/
static os_boolean
ADMXRC_GetCardInfo_P_2L(
  ADMXRC* devCtx)
{
  unsigned long data;
  os_uint32     data32;
  unsigned long i;
  PLX9080REGS*  plx;

  os_dbgprint(1, ("ADMXRC_GetCardInfo_P_2L: entered\n"));

  plx = devCtx->hw.regs.pci9080;

  devCtx->info.regsLimit = 0x100;

  switch (devCtx->id.pci.subDId) {
  case PCI_SUBDEVICE_ID_ADMXRC:
    {
      /*
      ** ADM-XRC or ADM-XRC-P
      **
      ** Location 0x36: board type
      */
      ADMXRC_ReadEprom(devCtx, 0x36, &data);
      if ((data == ADMXRC2_BOARD_ADMXRC) ||
	  (data == ADMXRC2_BOARD_ADMXRC_P)) {
	devCtx->info.card.BoardType = (ADMXRC2_BOARD_TYPE) data;
	devCtx->info.card.NumClock = 2;
	devCtx->info.card.NumDMAChan = 2;
      } else {
	/*
	** Board type must be recorded wrongly in EEPROM,
	** assume ADM-XRC.
	*/
	devCtx->info.card.BoardType = ADMXRC2_BOARD_ADMXRC;
	devCtx->info.card.NumClock = 2;
	devCtx->info.card.NumDMAChan = 2;
      }
      break;
    }

  case PCI_SUBDEVICE_ID_ADMXRC2_L:
    {
      /*
      ** ADM-XRC2-L
      **
      ** Don't use the board type from the EEPROM, as
      ** the PCI subsystem device ID for the ADM-XRC2-L
      ** is sufficient to correctly identify it.
      */
      devCtx->info.card.BoardType = ADMXRC2_BOARD_ADMXRC2_LITE;
      devCtx->info.card.NumClock = 2;
      devCtx->info.card.NumDMAChan = 2;
      break;
    }

  default:
    {
      os_dbgprint(1, ("ADMXRC_GetCardInfo_P_2L: unrecognised subsystem device ID: 0x%4.4lx\n",
		      (unsigned long) devCtx->id.pci.subDId));
      return FALSE;
    }
  }

  /*
  ** Location 0x37: modifications word
  **
  ** The modifications word is interpreted in a board
  ** specific manner.
  */
  devCtx->info.modifications = 0x0;
  ADMXRC_ReadEprom(devCtx, 0x37, &data);
  switch (devCtx->info.card.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
    {
      /*
      ** ADM-XRC/ADM-XRC-P:
      **
      ** Bit     Meaning
      **
      ** 0       1 => clock swap mod has been performed
      */
      if (data & 0x1) {
	devCtx->info.modifications |= ADMXRC_MOD_CLOCKSWAP;
      }
      break;
    }

  case ADMXRC2_BOARD_ADMXRC2_LITE:
    {
      /*
      ** ADM-XRC-II-L: no modifications currently defined
      */
      break;
    }

  default:
    {
      /*
      ** Just stop the compiler issuing a warning - we know we've already
      ** checked the board type for validity.
      */
    }
  }
  
  /* Location 0x38: clock oscillator type (ie. frequency) */
  ADMXRC_ReadEprom(devCtx, 0x38, &data);  
  devCtx->info.clockType = (ADMXRC_CLOCK_TYPE) data;
  
  /* Location 0x39: flash type (ie. device) */
  ADMXRC_ReadEprom(devCtx, 0x39, &data);  
  devCtx->info.flashType = (ADMXRC_FLASH_TYPE) data;
  
  /* Location 0x3B: card ID */
  ADMXRC_ReadEprom(devCtx, 0x3B, &data);
  devCtx->info.card.CardID = (ADMXRC2_CARDID) data;
  
  /* Location 0x3C: fpga type (ie. device) */
  ADMXRC_ReadEprom(devCtx, 0x3C, &data);
  devCtx->info.card.FPGAType = (ADMXRC2_FPGA_TYPE) data;
  
  /* Location 0x3D: serial number low 16 bits */
  /* Location 0x3E: serial number high 16 bits */
  ADMXRC_ReadEprom(devCtx, 0x3D, &data);
  devCtx->info.card.SerialNum = data;
  ADMXRC_ReadEprom(devCtx, 0x3E, &data);
  devCtx->info.card.SerialNum |= data << 16;
  
  /*
  ** Location 0x3F:
  **    low 8 bits = logic rev
  **    high 8 bits = board rev
  */
  ADMXRC_ReadEprom(devCtx, 0x3F, &data);
  devCtx->info.card.BoardRevision = (os_uint8) (data >> 8);
  devCtx->info.card.LogicRevision = (os_uint8) (data & 0xffU);
  
  /* Location 0x3A: SRAM presence info */
  ADMXRC_ReadEprom(devCtx, 0x3A, &data);
  switch (devCtx->info.card.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
    {
      /*
      ** ADM-XRC/ADM-XRC-P: get bank size/presence info, 4 banks
      **
      ** 4 banks of x36 SSRAM
      */
      os_boolean found = FALSE;
      
      devCtx->info.card.NumRAMBank = 4;
      if ((data & 0xf) == 0xf) {
	devCtx->info.card.RAMBanksFitted = 0x0;
	for (i = 0; i < 4; i++) {
	  devCtx->info.banks[i].Fitted = TRUE;
	  devCtx->info.banks[i].Type = ADMXRC2_RAM_ZBTFT;
	  devCtx->info.banks[i].Width = 36;
	  if (data & (1 << (i + 4))) {
	    /* 16Mbit */
	    devCtx->info.banks[i].Size = 512*1024;
	  } else {
	    /* 8Mbit */
	    devCtx->info.banks[i].Size = 256*1024;
	  }
	  devCtx->info.card.RAMBanksFitted |= (1 << i);
	}
	found = TRUE;
      }
      if (!found) {
	for (i = 0; i < 4; i++) {
	  devCtx->info.banks[i].Fitted = FALSE;
	  devCtx->info.banks[i].Type = 0;
	  devCtx->info.banks[i].Width = 36;
	  devCtx->info.banks[i].Size = 0;
	}
	devCtx->info.card.RAMBanksFitted = 0x0;
      }
      break;
    }

  case ADMXRC2_BOARD_ADMXRC2_LITE:
    {
      /*	
      ** ADM-XRC-IIL: get bank size/presence info
      **
      ** 4 banks of x18 SSRAM
      */
      os_uint16  bank;
      os_boolean fitted = FALSE; /* Shut the compiler up */
      
      devCtx->info.card.NumRAMBank = 4;
      for (i = 0; i < devCtx->info.card.NumRAMBank; i++) {
	bank = (data >> (i << 2)) & 0xf;
	switch (bank & 0x3) {
	case 0x0:
	  /* not fitted */
	  fitted = FALSE;
	  devCtx->info.banks[i].Type = 0;
	  break;
	case 0x1:
	  /* ZBT flow-through */
	  fitted = TRUE;
	  devCtx->info.banks[i].Type = ADMXRC2_RAM_ZBTFT;
	  break;
	case 0x2:
	  /* ZBT pipelined */
	  fitted = TRUE;
	  devCtx->info.banks[i].Type = ADMXRC2_RAM_ZBTP;
	  break;
	case 0x3:
	  /* ZBT flow-through + pipelined (selectable) */
	  fitted = TRUE;
	  devCtx->info.banks[i].Type = ADMXRC2_RAM_ZBTFT | ADMXRC2_RAM_ZBTP;
	  break;
	}
	devCtx->info.banks[i].Width = 18;
	if (!fitted) {
	  devCtx->info.banks[i].Size = 0;
	} else {
	  switch ((bank >> 2) & 0x3) {
	  case 0x0:
	    /* 8Mbit */
	    devCtx->info.banks[i].Size = 512*1024;
	    break;
	  case 0x1:
	    /* 16Mbit */
	    devCtx->info.banks[i].Size = 1024*1024;
	    break;
	  case 0x2:
	    /* 32Mbit */
	    devCtx->info.banks[i].Size = 2048*1024;
	    break;
	  case 0x3:
	    /* reserved */
	    devCtx->info.banks[i].Size = 0;
	    fitted = FALSE;
	    break;
	  }
	}
	devCtx->info.banks[i].Fitted = fitted;
	if (fitted) {
	  devCtx->info.card.RAMBanksFitted |= (1 << i);
	}
      }
      break;
    }

  default:
    {
      break;
    }
  }
  
  /*
  ** Find the bases & limits of the PLX spaces S0, S1
  */
  {
    data32 = os_read_register_32endian(&plx->LAS0BA);
    if (!(data32 & 0x1U)) {
      /* Space not enabled - we don't expect this */
      devCtx->info.spaces[0].LocalBase = 0x0U;
      devCtx->info.spaces[0].LocalSize = 0x0U;
    } else {
      devCtx->info.spaces[0].LocalBase = data32 & ~0xf;
      data32 = os_read_register_32endian(&plx->LAS0RR);
      devCtx->info.spaces[0].LocalSize = (~(data32 & ~0xf)) + 1;
    }
    devCtx->info.spaces[0].PhysicalBase = (unsigned long) devCtx->resources.raw.pci.base[2];
    devCtx->info.spaces[0].VirtualBase = NULL;
    devCtx->info.spaces[0].VirtualSize = 0;
       
    data32 = os_read_register_32endian(&plx->LAS1BA);
    if (!(data32 & 0x1)) {
      /* Space not enabled - we don't expect this */
      devCtx->info.spaces[1].LocalBase = 0x0U;
      devCtx->info.spaces[1].LocalSize = 0x0U;
    } else {
      devCtx->info.spaces[1].LocalBase = data32 & ~0xf;
      data32 = os_read_register_32endian(&plx->LAS1RR);
      devCtx->info.spaces[1].LocalSize = (~(data32 & ~0xf)) + 1;
    }
    devCtx->info.spaces[1].PhysicalBase = (unsigned long) devCtx->resources.raw.pci.base[3];
    devCtx->info.spaces[1].VirtualBase = NULL;
    devCtx->info.spaces[1].VirtualSize = 0;
    
    devCtx->info.card.NumSpace = 2;    
  }

  return TRUE;
}

static os_boolean
ADMXRC_GetCardInfo_2(
  ADMXRC* devCtx)
{
  unsigned long data;
  os_uint32     data32;
  unsigned long i;
  PLX9080REGS*  plx;

  os_dbgprint(1, ("ADMXRC_GetCardInfo_P_2L: entered\n"));

  plx = devCtx->hw.regs.pci9080;

  devCtx->info.regsLimit = 0x100;

  devCtx->info.card.BoardType = ADMXRC2_BOARD_ADMXRC2;
  devCtx->info.card.NumClock = 2;
  devCtx->info.card.NumDMAChan = 2;

  /*
  ** Location 0x36: modifications word
  **
  ** The modifications word is interpreted in a board
  ** specific manner.
  */
  ADMXRC_ReadEprom(devCtx, 0x36, &data);
  devCtx->info.modifications = 0x0; /* No modifications defined yet */

  /* Location 0x37: clock oscillator type (ie. frequency) */
  ADMXRC_ReadEprom(devCtx, 0x37, &data);
  devCtx->info.clockType = (ADMXRC_CLOCK_TYPE) data;
  
  /* Location 0x38: flash type (ie. device) */
  ADMXRC_ReadEprom(devCtx, 0x38, &data);
  devCtx->info.flashType = (ADMXRC_FLASH_TYPE) data;

  /* Location 0x3B: card ID */
  ADMXRC_ReadEprom(devCtx, 0x3B, &data);
  devCtx->info.card.CardID = (ADMXRC2_CARDID) data;
  
  /* Location 0x3C: fpga type (ie. device) */
  ADMXRC_ReadEprom(devCtx, 0x3C, &data);
  devCtx->info.card.FPGAType = (ADMXRC2_FPGA_TYPE) data;

  /* Location 0x3D: serial number low 16 bits */
  /* Location 0x3E: serial number high 16 bits */
  ADMXRC_ReadEprom(devCtx, 0x3D, &data);
  devCtx->info.card.SerialNum = data;
  ADMXRC_ReadEprom(devCtx, 0x3E, &data);
  devCtx->info.card.SerialNum |= data << 16;

  /*
  ** Location 0x3F:
  **    low 8 bits = logic rev
  **    high 8 bits = board rev
  */
  ADMXRC_ReadEprom(devCtx, 0x3F, &data);
  devCtx->info.card.BoardRevision = (os_uint8) (data >> 8);
  devCtx->info.card.LogicRevision = (os_uint8) (data & 0xFF);
  
  /* Locations 0x39 & 0x3A: SRAM presence info */
  ADMXRC_ReadEprom(devCtx, 0x39, &data);
  data32 = data;
  ADMXRC_ReadEprom(devCtx, 0x3A, &data);
  data32 |= data << 16U;
  {
    /*
    ** ADM-XRC2: get bank size/presence info, 6 banks
    */
    unsigned long bank;
    os_boolean    fitted = FALSE;
    
    devCtx->info.card.NumRAMBank = 6;
    devCtx->info.card.RAMBanksFitted = 0;
    for (i = 0; i < devCtx->info.card.NumRAMBank; i++) {
      bank = (data32 >> (i << 2)) & 0xf;
      switch (bank & 0x3) {
      case 0x0:
	/* not fitted */
	fitted = FALSE;
	devCtx->info.banks[i].Type = 0;
	break;
      case 0x1:
	/* ZBT flow-through */
	fitted = TRUE;
	devCtx->info.banks[i].Type = ADMXRC2_RAM_ZBTFT;
	break;
      case 0x2:
	/* ZBT pipelined */
	fitted = TRUE;
	devCtx->info.banks[i].Type = ADMXRC2_RAM_ZBTP;
	break;
      case 0x3:
	/* ZBT flow-through + pipelined (selectable) */
	fitted = TRUE;
	devCtx->info.banks[i].Type = ADMXRC2_RAM_ZBTFT | ADMXRC2_RAM_ZBTP;
	break;
      }
      devCtx->info.banks[i].Width = (devCtx->info.card.FPGAType == ADMXRC2_FPGA_2V3000) ? 32 : 36;
      if (!fitted) {
	devCtx->info.banks[i].Size = 0;
      } else {
	switch ((bank >> 2) & 0x3) {
	case 0x0:
	  /* 8Mbit */
	  devCtx->info.banks[i].Size = 256*1024;
	  break;
	case 0x1:
	  /* 16Mbit */
	  devCtx->info.banks[i].Size = 512*1024;
	  break;
	case 0x2:
	  /* 32Mbit */
	  devCtx->info.banks[i].Size = 1024*1024;
	  break;
	case 0x3:
	  /* 64Mbit */
	  devCtx->info.banks[i].Size = 2048*1024;
	  fitted = FALSE;
	  break;
	}
      }
      devCtx->info.banks[i].Fitted = fitted;
      if (fitted) {
	devCtx->info.card.RAMBanksFitted |= (1 << i);
      }
    }
  }

  /*
  ** Find the bases & limits of the PLX spaces S0, S1
  */
  {
    data32 = os_read_register_32endian(&plx->LAS0BA);
    if (!(data32 & 0x1U)) {
      /* Space not enabled - we don't expect this */
      devCtx->info.spaces[0].LocalBase = 0x0U;
      devCtx->info.spaces[0].LocalSize = 0x0U;
    } else {
      devCtx->info.spaces[0].LocalBase = data32 & ~0xf;
      data32 = os_read_register_32endian(&plx->LAS0RR);
      devCtx->info.spaces[0].LocalSize = (~(data32 & ~0xf)) + 1;
    }
    devCtx->info.spaces[0].PhysicalBase = (unsigned long) devCtx->resources.raw.pci.base[2];
    devCtx->info.spaces[0].VirtualBase = NULL;
    devCtx->info.spaces[0].VirtualSize = 0;
       
    data32 = os_read_register_32endian(&plx->LAS1BA);
    if (!(data32 & 0x1)) {
      /* Space not enabled - we don't expect this */
      devCtx->info.spaces[1].LocalBase = 0x0U;
      devCtx->info.spaces[1].LocalSize = 0x0U;
    } else {
      devCtx->info.spaces[1].LocalBase = data32 & ~0xf;
      data32 = os_read_register_32endian(&plx->LAS1RR);
      devCtx->info.spaces[1].LocalSize = (~(data32 & ~0xf)) + 1;
    }
    devCtx->info.spaces[1].PhysicalBase = (unsigned long) devCtx->resources.raw.pci.base[3];
    devCtx->info.spaces[1].VirtualBase = NULL;
    devCtx->info.spaces[1].VirtualSize = 0;
    
    devCtx->info.card.NumSpace = 2;    
  }

  return TRUE;
}

os_boolean
ADMXRC_SetupCardInfo(
  ADMXRC* devCtx)
{
  os_boolean found;

  os_dbgprint(1, ("ADMXRC_SetupCardInfo: entered\n"));
  
  switch (devCtx->id.pci.subDId) {
  case PCI_SUBDEVICE_ID_ADMXRC:
  case PCI_SUBDEVICE_ID_ADMXRC2_L:
    devCtx->info.eepromType = ADMXRC_EEPROM_9346;
    found = ADMXRC_GetCardInfo_P_2L(devCtx);
    break;
  case PCI_SUBDEVICE_ID_ADMXRC2:
    devCtx->info.eepromType = ADMXRC_EEPROM_9366;
    found = ADMXRC_GetCardInfo_2(devCtx);
    break;
  default:
    os_dbgprint(0, ("ADMXRC_SetupCardInfo: unrecognised hardware ID: 0x%4.4lx%4.4lx 0x%4.4lx%4.4lx\n",
		    (unsigned long) devCtx->id.pci.vId,
		    (unsigned long) devCtx->id.pci.dId,
		    (unsigned long) devCtx->id.pci.subVId,
		    (unsigned long) devCtx->id.pci.subDId));
    /* Shouldn't happen! */
    found = FALSE;
  }
  
  os_dbgprint(1, ("ADMXRC_SetupCardInfo: CardID = 0x%lx(%ld)\n",
		  (unsigned long) devCtx->info.card.CardID,
		  (unsigned long) devCtx->info.card.CardID));

  os_dbgprint(1, ("ADMXRC_SetupCardInfo: successful\n"));

  return found;
}
