#ifndef _ADATA_DEVICE_H
#define _ADATA_DEVICE_H

/*
** Device structure definition for the ADM-XRC series driver
*/

#include "platform.h"
#include "device_defs.h"
#include "plx.h"
#include "admxrc/struct.h"
#include "admxrc2/status.h"
#include "admxrc2/struct.h"

/*
** Declare some structures now so we can forward-reference them
*/
struct _ADMXRC;
typedef struct _ADMXRC ADMXRC;
struct _ADMXRC_DMAQ;
typedef struct _ADMXRC_DMAQ ADMXRC_DMAQ;
struct _ADMXRC_DMANODE;
typedef struct _ADMXRC_DMANODE ADMXRC_DMANODE;
struct _ADMXRC_DMAABTCTX;
typedef struct _ADMXRC_DMAABTCTX ADMXRC_DMAABTCTX;

typedef enum _ADMXRC_EEPROM_TYPE {
  ADMXRC_EEPROM_9346 = 0,
  ADMXRC_EEPROM_9356 = 1,
  ADMXRC_EEPROM_9366 = 2
} ADMXRC_EEPROM_TYPE;

typedef enum _ADMXRC_FLASH_TYPE {
  ADMXRC_FLASH_SST    = 0,
  ADMXRC_FLASH_MICRON = 1,
  ADMXRC_FLASH_CFI    = 2
} ADMXRC_FLASH_TYPE;

/*
** A structure representing a region of memory that has been
** locked down as a target for DMA transfers.
*/
typedef struct _ADMXRC_DMATGT {
  OS_USBD*      usbd;   /* description of mapped & locked region      */
  unsigned long length;	/* size of mapped & locked region             */
  unsigned long	in_use; /* (in_use > 0) => cannot be freed, is in use */
} ADMXRC_DMATGT;

/*
** A DMA completion callback routine - called when a DMA transfer
** is completed.
*/
typedef void (ADMXRC_DMACALLBACK) (ADMXRC* devCtx, ADMXRC_DMANODE* node);

/*
** A DMA queue, for queuing up DMA transfers
*/
struct _ADMXRC_DMAQ {
  int             length;
  ADMXRC_DMANODE* head;
  ADMXRC_DMANODE* tail;
};

/*
** A context structure used when aborting DMA transfers
*/
struct _ADMXRC_DMAABTCTX {
  /* initialised before call to atomic function */
  ADMXRC*       devCtx;
  unsigned char channel;

  /* filled in by call to atomic function */
  os_boolean    success;
};

/*
** A structure for controlling a DMA transfer; usually part of
** a queue of DMA transfers.
*/
struct _ADMXRC_DMANODE {
  /* Client code must not touch these members */
  struct {
    ADMXRC_DMANODE*     prev;
    ADMXRC_DMANODE*     next;
    unsigned char       state;      /* 1 => queued, 2 => active, 3 => completed */
    unsigned char       activechan; /* Valid when state is 'active on hardware' */
  } system;

  /* Client code must initialize these members */
  struct {
    void*               context;
    ADMXRC_DMATGT*      target;
    ADMXRC_DMACALLBACK* callback;
    os_uint32           mode;
    unsigned long       length;   /* WARNING - decremented during DMA */
    os_uint32           local;    /* WARNING - incremented during DMA */
    unsigned long       pos;      /* WARNING - incremented during DMA */
    unsigned char       direction;
    unsigned char       queue;
  } client;

  /* Client code may inspect these members on completion */
  struct {
    OS_STATUS           status;
  } status;
};

/*
** A structure for controlling an FPGA configuration operation that
** uses DMA.
*/
typedef struct _ADMXRC_LOADFPGADMA {
  ADMXRC_DMANODE      dmanode;
  OS_EVENT            event;
} ADMXRC_LOADFPGADMA;

struct _ADMXRC {
  /* Saved from onCreateDevice */
  OS_DRIVER_OBJECT*     drvObj;
  OS_DEVICE_OBJECT*     devObj;
  OS_DEVICE_DESC        id;

  /* Resources given in onStartDevice */
  OS_DEVICE_RESOURCES   resources;

  /* Mapped/attached resources */
  OS_ATTACHED_IRQ*      irqObj;
  void*                 bar[4];

  /* Hardware mapped into kernel space */
  struct {
    union {
      void*             untyped;
      PLX9080REGS*      pci9080;
      PLX9656REGS*      pci9656;
    } regs;
    void*               fpga;
    CONTROLXRC*         control;
    void*               flash;
  } hw;

  /* Shadows of important registers */
  struct {
    os_uint16           dmacsr;
  } shadow;

  struct {
    /* Are interrupts enabled or disabled? */
    os_boolean            enabled;
    OS_DPC*               fpgaintDpc;
    OS_DPC*               dmaintDpc[ADMXRC_NUM_DMA_CHANNEL_MAX];
  } interrupts;

  /* Tracks the open handle to this device that has 'modify access' */
  struct {
    OS_USER_CLIENT      handle;
    OS_SPINLOCK         lock;
  } owner;

  /* DMA resources and data structures */
  struct {
    OS_SPINLOCK         lock;
    OS_LCACHE*          ctlPool;
    struct {
      struct {
	void*           real;
	PLXDESC*        aligned;
	os_uint32       pci[ADMXRC_MAX_PLXDESC_PER_CHANNEL]; 
      } descriptors;
      ADMXRC_DMANODE*   active;
      unsigned long     activeCnt;
    } channel[ADMXRC_NUM_DMA_CHANNEL_MAX];
    ADMXRC_DMAQ         queue[ADMXRC_NUM_DMA_CHANNEL_MAX + 1];
  } dma;

  /* Memory regions mapped into user space */
  struct {
    OS_MAPPABLE_REGION* regions[6];
    int                 count;
  } userMapped;

  /* Card information, which user-mode threads can read */
  struct {
    ADMXRC2_CARD_INFO   card;
    ADMXRC2_SPACE_INFO  spaces[ADMXRC_NUM_SPACE_MAX];
    ADMXRC2_BANK_INFO   banks[ADMXRC_NUM_BANK_MAX];
    os_uint32           regsLimit;
    ADMXRC_EEPROM_TYPE  eepromType;
    ADMXRC_CLOCK_TYPE   clockType;
    ADMXRC_FLASH_TYPE   flashType;
    unsigned long       modifications; /* format is board-specific */
  } info;

  /* FPGA information and control */
  struct {
    OS_LCACHE*          ctlPool;
    OS_SPINLOCK         lock;
    os_boolean          busy;
  } fpga;

  /* Interrupt waiting */
  struct {
    os_boolean          signalled;
    struct {
      OS_IO_REQUEST*    req;
      os_boolean        completing;
    } waiting;
    struct {
      unsigned long     ticks;
      os_boolean        never;
    } expire;
    OS_SPINLOCK         lock;
  } intwait;

  /* Clock generator programming */
  struct {
    os_uint32           clockword;
  } setclock;

  /* DMA descriptors */
  struct {
    ADMXRC_DMATGT**     table;     /* Maps DMADESCs to ADMXRC_DMATGTs */
    int*                freestack; /* Stack of free table indices     */
    int                 top;       /* Stack pointer for 'freestack'   */
    OS_SPINLOCK         lock;
  } dmadesc;
};

#endif

