/*
** dma.c - DMA IOCTL handlers
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "dma.h"
#include "dmadesc.h"
#include "dmacore.h"

/*
** This callback function is attached to the DMANODEs created by
** the DoDMAImmediate IOCTL handler.
**
** May be called in no thread context.
*/
static void
doDMACallback(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node)
{
  OS_IO_REQUEST* req;
  ADMXRC_DMATGT* dmatgt;
  OS_STATUS      status;

  req = (OS_IO_REQUEST*) node->client.context;
  dmatgt = node->client.target;
  status = node->status.status;

  if (os_set_cancel_routine(req, NULL)) {
    /* Return the DMANODE to the lookaside cache */
    os_free_lcache(devCtx->dma.ctlPool, node);
  
    /*
    ** Unlock the user buffer - is this safe to do outside a thread context?
    **
    ** Consider moving this to a thread context so it doesn't hold things up
    ** if it turns out to be a time-consuming operation.
    */
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  
    /* Complete the IO request with the status from the DMANODE */
    req->status.status = status;
    os_complete_request(req);
  }
}

/*
** This callback function is attached to the IO request that 'owns' a DMANODE
**
** May be called in no thread context.
*/
static void
doDMACancelReq(
  OS_DEVICE_OBJECT* devObj,
  OS_IO_REQUEST*    req)
{
  ADMXRC*           devCtx;
  ADMXRC_DMANODE*   node;
  ADMXRC_DMATGT*    dmatgt;

  devCtx = (ADMXRC*) os_get_device_context(devObj);
  node = (ADMXRC_DMANODE*) req->user.context;

  if (ADMXRC_CancelDMA(devCtx, node)) {
    dmatgt = node->client.target;

    /* Return the DMANODE to the lookaside cache */
    os_free_lcache(devCtx->dma.ctlPool, node);
    
    /*
    ** Unlock the user buffer - is this safe to do outside a thread context?
    **
    ** Consider moving this to a thread context so it doesn't hold things up
    ** if it turns out to be a time-consuming operation.
    */
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  
    /* Complete the IO request */
    req->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
    os_complete_request(req);
  }
}

/*
** Handle ADMXRC_IOCTL_DODMA
*/
OS_STATUS
ADMXRC_DoDMA(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMA* iptr;
  ADMXRC_DMANODE*      dmanode;
  ADMXRC_DMATGT*       dmatgt;
  OS_STATUS            status;
  os_uint32            local;
  unsigned long        len;
  os_uint32            mode;
  unsigned long        flags;
  unsigned long        channel;
  unsigned long        dir;
  unsigned long        offset;
  unsigned char        queue;
  os_boolean           cleanup = FALSE;
  os_boolean           canstart;
  OS_SPINLOCK_FLAGS    f;

  iptr = (ADMXRC_IOCTLS_DODMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC_PCI2LOCAL && dir != ADMXRC_LOCAL2PCI) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = iptr->in.local;
  len = iptr->in.length;
  if (len == 0) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (mode & ADMXRC_DMAMODE_FIXEDLOCAL) {
    switch (mode & 0x3)
      {
      case 0:
	if (local > 0xffffff) {
	  return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
	}
      case 1:
	if (local > 0xfffffe) {
	  return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
	}
      case 2:
      case 3:
	if (local > 0xfffffc) {
	  return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
	}
      }
  } else { 
    if (local >= 0x1000000) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
    if ((local + len) > 0x1000000) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
  }
  
  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.card.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  dmanode = (ADMXRC_DMANODE*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
  }

  /* Lock down the user buffer and obtain an MDL for it */
  dmatgt = ADMXRC_GetDMATarget(devCtx, iptr->in.dmadesc);
  if (dmatgt == NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
    return OS_MAKE_STATUS(ADMXRC_INVALID_DMADESC);
  }

  /* Validate the offset and size parameters */
  offset = iptr->in.offset;
  if (offset >= dmatgt->length || (offset + len) > dmatgt->length) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Add bits to PCI9080/PCI9656 DMA mode word that are mandatory for this driver */
  mode |=
    (1 << 10) |                             /* done interrupt enable       */
    (ADMXRC_ALLOW_WRITE_INVALIDATE << 13) | /* write and invalidate cycles */
    (1 << 17);                              /* route interrupt to PCI bus  */

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.card.NumDMAChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) req;
  dmanode->client.target = dmatgt;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = len;
  dmanode->client.local = local;
  dmanode->client.pos = os_get_usbd_page_offset(dmatgt->usbd) + offset;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMA: enqueuing DMANODE 0x%p dmadesc=%ld offset=%ld len=%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (unsigned long) iptr->in.dmadesc,
		  (unsigned long) offset,
		  (unsigned long) len,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  /* Queue up the DMA (it might begin on hardware immediately) */
  req->status.status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_acquire_spinlock(&devCtx->dma.lock, &f);
  canstart = ADMXRC_TestQueueDMA(devCtx, dmanode);
  if ((flags & ADMXRC_DMAFLAG_DONOTQUEUE) && !canstart) {
    cleanup = TRUE;
    status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  } else {
    if (os_set_cancel_routine(req, doDMACancelReq)) {
      ADMXRC_EnqueueDMA(devCtx, dmanode, canstart);
      status = OS_STATUS_PENDING;
    } else {
      cleanup = TRUE;
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
    }
  }
  os_release_spinlock(&devCtx->dma.lock, f);

  if (cleanup) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }

  return status;
}

/*
** This callback function is attached to the DMANODEs created by
** the DoDMAImmediate IOCTL handler.
**
** May be called in no thread context.
*/
static void
doDMAImmediateCallback(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node)
{
  OS_IO_REQUEST* req;
  ADMXRC_DMATGT* dmatgt;
  OS_STATUS      status;

  req = (OS_IO_REQUEST*) node->client.context;
  dmatgt = node->client.target;
  status = node->status.status;

  if (os_set_cancel_routine(req, NULL)) {
    /* Return the DMANODE to the lookaside cache */
    os_free_lcache(devCtx->dma.ctlPool, node);
  
    /*
    ** Unlock the user buffer - is this safe to do outside a thread context?
    **
    ** Consider moving this to a thread context so it doesn't hold things up
    ** if it turns out to be a time-consuming operation.
    */
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  
    /* Complete the IO request with the status from the DMANODE */
    req->status.status = status;
    os_complete_request(req);
  }
}

/*
** This callback function is attached to the IO request that 'owns' a DMANODE
**
** May be called in no thread context.
*/
static void
doDMAImmediateCancelReq(
  OS_DEVICE_OBJECT* devObj,
  OS_IO_REQUEST*    req)
{
  ADMXRC*           devCtx;
  ADMXRC_DMANODE*   node;
  ADMXRC_DMATGT*    dmatgt;

  devCtx = (ADMXRC*) os_get_device_context(devObj);
  node = (ADMXRC_DMANODE*) req->user.context;

  if (ADMXRC_CancelDMA(devCtx, node)) {
    dmatgt = node->client.target;

    /* Return the DMANODE to the lookaside cache */
    os_free_lcache(devCtx->dma.ctlPool, node);
    
    /*
    ** Unlock the user buffer - is this safe to do outside a thread context?
    **
    ** Consider moving this to a thread context so it doesn't hold things up
    ** if it turns out to be a time-consuming operation.
    */
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  
    /* Complete the IO request */
    req->status.status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
    os_complete_request(req);
  }
}

/*
** Handle ADMXRC_IOCTL_DODMAIMMEDIATE
*/
OS_STATUS
ADMXRC_DoDMAImmediate(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMAIMMEDIATE*	iptr;
  ADMXRC_DMANODE*               dmanode;
  ADMXRC_DMATGT*                dmatgt;
  OS_STATUS                     status;
  os_uint32                     local;
  unsigned long                 len;
  os_uint32                     mode;
  unsigned long                 flags;
  unsigned long                 channel;
  unsigned long                 dir;
  unsigned char                 queue;
  os_boolean                    cleanup = FALSE;
  os_boolean                    canstart;
  OS_SPINLOCK_FLAGS             f;

  iptr = (ADMXRC_IOCTLS_DODMAIMMEDIATE*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC_PCI2LOCAL && dir != ADMXRC_LOCAL2PCI) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = iptr->in.local;
  len = iptr->in.length;
  if (len == 0) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (mode & ADMXRC_DMAMODE_FIXEDLOCAL) {
    switch (mode & 0x3)
      {
      case 0:
	if (local > 0xffffff) {
	  return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
	}
      case 1:
	if (local > 0xfffffe) {
	  return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
	}
      case 2:
      case 3:
	if (local > 0xfffffc) {
	  return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
	}
      }
  } else { 
    if (local >= 0x1000000) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
    if ((local + len) > 0x1000000) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
  }
  
  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.card.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  dmanode = (ADMXRC_DMANODE*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
  }

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(devCtx, iptr->in.buffer, iptr->in.length, &dmatgt);
  if (!OS_SUCCESS(status)) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
    return status;
  }

  /* Add bits to PCI9080/PCI9656 DMA mode word that are mandatory for this driver */
  mode |=
    (1 << 10) |                             /* done interrupt enable       */
    (ADMXRC_ALLOW_WRITE_INVALIDATE << 13) | /* write and invalidate cycles */
    (1 << 17);                              /* route interrupt to PCI bus  */

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.card.NumDMAChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) req;
  dmanode->client.target = dmatgt;
  dmanode->client.callback = doDMAImmediateCallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = len;
  dmanode->client.local = local;
  dmanode->client.pos = os_get_usbd_page_offset(dmatgt->usbd);
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMAImmediate: enqueuing DMANODE 0x%p buffer=0x%p len=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (void*) iptr->in.buffer,
		  (unsigned long) len,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  /* Queue up the DMA (it might begin on hardware immediately) */
  req->status.status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_acquire_spinlock(&devCtx->dma.lock, &f);
  canstart = ADMXRC_TestQueueDMA(devCtx, dmanode);
  if ((flags & ADMXRC_DMAFLAG_DONOTQUEUE) && !canstart) {
    cleanup = TRUE;
    status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  } else {
    if (os_set_cancel_routine(req, doDMAImmediateCancelReq)) {
      ADMXRC_EnqueueDMA(devCtx, dmanode, canstart);
      status = OS_STATUS_PENDING;
    } else {
      cleanup = TRUE;
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
    }
  }
  os_release_spinlock(&devCtx->dma.lock, f);

  if (cleanup) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }

  return status;
}
