/*
** dmacore.c
**
** Functions to support DMA transfers. Exported functions are:
**
**   ADMXRC_TestQueueDMA           Checks whether a DMA transfer can
**                                 begin immediately and sets up
**                                 certain members in the node.
**
**   ADMXRC_EnqueueDMA             Queues or begins a DMA transfer
**
**   ADMXRC_CancelDMA              Cancels a DMA transfer
**
**   ADMXRC_CleanupDMA             Called when shutting down the
**                                 device; aborts all DMA transfers
**                                 and completes them with a failure
**                                 status.
**
**
** Primary datatype is the ADMXRC_DMANODE, which is a control structure
** for a single DMA operation.
**
** A DMA node's state is as follows:
**
**   1     => in a DMA queue (not yet active on the hardware)
**   2     => active on the hardware
**   3     => completed by hardware
**   other => state is invalid
**
** Each DMA node has a callback function associated with it, called
** when it is completed. The callback function may be invoked in an
** arbitrary context.
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "dma.h"
#include "dmadesc.h"

/*
** Removes a DMANODE from (anywhere in) a queue.
**
** Must be called while holding the DMA spinlock
*/
static void
removeDMA(
  ADMXRC_DMAQ*    q,
  ADMXRC_DMANODE* node)
{
  ADMXRC_DMANODE* next;
  ADMXRC_DMANODE* prev;

  next = node->system.next;
  prev = node->system.prev;
  if (next != NULL) {
    next->system.prev = prev;
  } else {
    q->tail = prev;
  }
  if (prev != NULL) {
    prev->system.next = next;
  } else {
    q->head = next;
  }
  q->length--;
}

/*
** Adds a DMANODE to the tail of a queue.
**
** Must be called while holding the DMA spinlock
*/
static void
addDMA(
  ADMXRC_DMAQ*    q,
  ADMXRC_DMANODE* node)
{
  ADMXRC_DMANODE* head;
  ADMXRC_DMANODE* tail;
  
  head = q->head;
  tail = q->tail;
  if (head == NULL) {
    q->tail = q->head = node;
    node->system.next = NULL;
    node->system.prev = NULL;
  } else {
    tail->system.next = node;
    q->tail = node;
    node->system.prev = tail;
    node->system.next = NULL;
  }
  q->length++;
}

/*
** Helper function to map a portion of a user buffer into a set
** of PLX DMA descriptors. One page is always mapped into one
** descriptor, even if the pages are contiguous in physical memory.
*/
static unsigned long
buildDMADescs(
  OS_USBD*       usbd,
  os_uint32*     descPhys,
  PLXDESC*       descs,
  os_uint32      local,        /* local bus address */
  os_uint32      descflags,    /* PLXDESC.link flags */
  unsigned long  startpage,    /* Start page index within USBD */
  unsigned long  startoffset,  /* Offset within start page */ 
  unsigned long  maxpages,     /* Max. no of pages to map */
  unsigned long  maxbytes,     /* Max. no of bytes to map */
  os_boolean     fixedlocal,   /* TRUE if local bus address should be incremented */
  unsigned long* bytesmapped)  /* Returns number of bytes mapped */
{
  unsigned long npage = 0;
  unsigned long nbytes = 0;
  os_uint32     pci;
  unsigned long count;
  PLXDESC*      last = NULL;

  while (maxpages && maxbytes) {
    if (last != NULL) {
      last->link |= cpu_to_le32(*++descPhys);
    }
      
    pci = os_get_usbd_page_addr(usbd, startpage) + startoffset;
    count = os_page_size - startoffset;
    startoffset = 0;
      
    if (count > maxbytes) {
      count = maxbytes;
    }
    
    descs->pci   = cpu_to_le32(pci);
    descs->local = cpu_to_le32(local);
    descs->count = cpu_to_le32(count);
    descs->link  = cpu_to_le32(descflags);
    last = descs;
    descs++;
    
    maxpages--;
    maxbytes -= count;
    startpage++;
    nbytes += count;
    npage++;
    if (!fixedlocal) {
      local += count;
    }
  }
  last->link |= cpu_to_le32(PLXDESCF_ENDCHAIN | PLXDESCF_INTERRUPT);
  
  *bytesmapped = nbytes;
  return npage;
}

/*
** This function builds the PLXDESCs that describe a single chunk
** of a user requested DMA operation, and then kicks off a PLX DMA
** transfer using those descriptors.
**
** Does not use chained mode.
**
** MUST be called while holding the appropriate spinlock.
*/
static void
doDMAChunkNonChained(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node,
  unsigned char   chan)
{
  unsigned long        npage;
  unsigned long        nbytes;
  unsigned long        size = node->client.length;
  unsigned long        mode = node->client.mode;
  ADMXRC_DMA_DIRECTION dir = node->client.direction;
  os_boolean           fixed;
  os_uint16            dmacsr, mask;
  
  fixed = (mode & ADMXRC_DMAMODE_FIXEDLOCAL) != 0;
  
  /* Build a single DMA descriptor */
  npage = buildDMADescs(node->client.target->usbd,
			devCtx->dma.channel[chan].descriptors.pci,
			devCtx->dma.channel[chan].descriptors.aligned,
			node->client.local,
			(dir == ADMXRC_PCI2LOCAL) ? 0x1 : 0x9,
			node->client.pos >> os_page_shift,
			node->client.pos & os_page_mask_low,
			1,
			size,
			fixed,
			&nbytes);
  
  mask = chan ? 0x0200U : 0x0002U;
  dmacsr = os_read_register_16endian(&devCtx->hw.regs.pci9080->DMACSR);
  if (dmacsr & (mask << 3)) {
    os_dbgprint(6, ("doDMAChunkNonChained: starting DMA on chan %ld for DMA node 0x%p, mode=0x%lx\n",
		    (unsigned long) chan, (void*) node, (unsigned long) node->client.mode));
    
    /* Set transfer mode (doesn't add in chained mode bit) */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].MODE,
			       node->client.mode);
    /* Set PCI address */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].PADR,
			       le32_to_cpu(devCtx->dma.channel[chan].descriptors.aligned[0].pci));
    /* Set local bus address */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].LADR,
			       le32_to_cpu(devCtx->dma.channel[chan].descriptors.aligned[0].local));
    /* Set transfer length */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].SIZE,
			       le32_to_cpu(devCtx->dma.channel[chan].descriptors.aligned[0].count));
    /* Set next descriptor pointer */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].DPR,
			       le32_to_cpu(devCtx->dma.channel[chan].descriptors.aligned[0].link));
    /* Strobe the start bit */
    os_write_register_16endian(&devCtx->hw.regs.pci9080->DMACSR,
			       (os_uint16) (mask | devCtx->shadow.dmacsr));
  } else {
    os_dbgprint(0, ("*** doDMAChunkNonChained: DMA chan %ld not idle for DMA node 0x%p\n",
		    (unsigned long) chan, (void*) node));
  }

  node->client.pos += nbytes;
  node->client.length -= nbytes;
  if (!fixed) {
    node->client.local += nbytes;
  }  
}

/*
** This function builds the PLXDESCs that describe a single chunk
** of a user requested DMA operation, and then kicks off a PLX9080 DMA
** transfer using those descriptors.
**
** Uses chained mode.
**
** MUST be called while holding the appropriate spinlock.
*/
static void
doDMAChunkChained(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node,
  unsigned char   chan)
{
  unsigned long        npage;
  unsigned long        nbytes;
  unsigned long        size = node->client.length;
  unsigned long        mode = node->client.mode;
  ADMXRC_DMA_DIRECTION dir = node->client.direction;
  os_boolean           fixed;
  os_uint16            dmacsr, mask;
  
  fixed = (mode & ADMXRC_DMAMODE_FIXEDLOCAL) != 0;
  
  /* Build the DMA descriptors */
  npage = buildDMADescs(node->client.target->usbd,
			devCtx->dma.channel[chan].descriptors.pci,
			devCtx->dma.channel[chan].descriptors.aligned,
			node->client.local,
			(dir == ADMXRC_PCI2LOCAL) ? 0x1 : 0x9,
			node->client.pos >> os_page_shift,
			node->client.pos & (os_page_size - 1),
			ADMXRC_MAX_PLXDESC_PER_CHANNEL,
			size,
			fixed,
			&nbytes);
  
  if (os_dbglevel(9)) {
    unsigned long i;

    os_dbgprint(0, ("doDMAChunkChained: Dump of PLXDESCs:\n"));
    os_dbgprint(0, ("doDMAChunkChained: ### DESCVIRT-------- DESCPCI- [PCI----- LOCAL--- COUNT--- LINK----]\n"));
    for (i = 0; i < npage; i++) {
      os_dbgprint(0, ("%3.3ld %16.16lx %8.8lx [%8.8lx %8.8lx %8.8lx %8.8lx]\n",
		      (unsigned long) i,
		      (unsigned long) &devCtx->dma.channel[chan].descriptors.aligned[i],
		      (unsigned long) devCtx->dma.channel[chan].descriptors.pci[i],
		      (unsigned long) devCtx->dma.channel[chan].descriptors.aligned[i].pci,
		      (unsigned long) devCtx->dma.channel[chan].descriptors.aligned[i].local,
		      (unsigned long) devCtx->dma.channel[chan].descriptors.aligned[i].count,
		      (unsigned long) devCtx->dma.channel[chan].descriptors.aligned[i].link));
    }
  }
  
  mask = chan ? 0x0200U : 0x0002U;
  dmacsr = os_read_register_16endian(&devCtx->hw.regs.pci9080->DMACSR);
  if (dmacsr & (mask << 3)) {
    os_dbgprint(6, ("doDMAChunkChained: starting DMA on chan %ld for DMA node 0x%p, mode=0x%lx\n",
		    (unsigned long) chan, (void*) node, (unsigned long) node->client.mode));
    
    /* Set current transfer mode (adds in chained mode bit) */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].MODE,
			       node->client.mode | (1U << 9));
    /* Set current transfer size to 0 */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].SIZE, 
			       0);
    /* Set the descriptor pointer */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->DMA[chan].DPR,
			       devCtx->dma.channel[chan].descriptors.pci[0] | 0x1);
    /* Strobe the start bit */
    os_write_register_16endian(&devCtx->hw.regs.pci9080->DMACSR,
			       (os_uint16) (mask | devCtx->shadow.dmacsr));
  } else {
    os_dbgprint(0, ("*** doDMAChunkChained: DMA chan %ld not idle for DMA node 0x%p\n",
		    (unsigned long) chan, (void*) node));
  }

  node->client.pos += nbytes;
  node->client.length -= nbytes;
  if (!fixed) {
    node->client.local += nbytes;
  }
}

/*
** This function builds the PLXDESCs that describe a single chunk
** of a user requested DMA operation, and then kicks off a PLX9080 DMA
** transfer using those descriptors.
**
** MUST be called while holding the appropriate spinlock.
*/
static void
doDMAChunk(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node,
  unsigned char   chan)
{
  if (ADMXRC_ALLOW_DMA_CHAINED_MODE) {
    doDMAChunkChained(devCtx, node, chan);
  } else {
    doDMAChunkNonChained(devCtx, node, chan);
  }
}

/*
** Critical section that aborts a DMA transfer on a PLX DMA
** channel. Since it modifies the PLX.DMACSR[0] bit, it must
** not run concurrently with the interrupt handler.
*/
static void*
abortDMAOnHardwareCritical(
  void* arg)
{
  ADMXRC_DMAABTCTX* context = (ADMXRC_DMAABTCTX*) arg;
  ADMXRC*	    devCtx = context->devCtx;
  int               chan = (int) context->channel;
  os_uint16         mask;
  os_uint16         dmacsr;
  os_uint16         val;
  os_boolean        success = TRUE;

  mask = chan ? 0x0100 : 0x0001;
  
  /* First, disable the DMA channel */
  dmacsr = devCtx->shadow.dmacsr & ~mask;
  os_write_register_16endian(&devCtx->hw.regs.pci9080->DMACSR, dmacsr);

  /* Now check if DMA channel is active or not */
  val = os_read_register_16endian(&devCtx->hw.regs.pci9080->DMACSR);
  if (val & (mask << 4)) {
    /* DMA is not active so must have already finished */
  } else {
    /* DMA is active - abort it */
    os_write_register_16endian(&devCtx->hw.regs.pci9080->DMACSR, (os_uint16) (dmacsr | (mask << 2)));
    /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
    os_delay_busy_us(50);
    /* Did we succeed in aborting it? */
    val = os_read_register_16endian(&devCtx->hw.regs.pci9080->DMACSR);
    if (val & (mask << 4)) {
      /* Yes, was aborted ok */
      os_dbgprint(3, ("abortDMACritical: aborted DMA on channel %ld\n", (unsigned long) chan));
    } else {
      /* No! Not much we can do now, except restore things */
      os_dbgprint(0, ("*** abortDMACritical: failed to abort DMA on channel %ld!\n", (unsigned long) chan));
      success = FALSE;
    }
  }
  
  /*
  ** Reenable the DMA channel and clear the DMA interrupt that will have
  ** occurred if we aborted.
  */
  dmacsr |= mask | (mask << 3);
  os_write_register_16endian(&devCtx->hw.regs.pci9080->DMACSR, dmacsr);
  
  context->success = success;

  return NULL;
}

/*
** Called to abort DMA transfer that is active on the hardware.
**
** Returns TRUE if DMA transfer was successfully aborted, otherwise FALSE.
*/
static os_boolean
abortDMAOnHardware(
  ADMXRC*       devCtx,
  unsigned char chan)
{
  ADMXRC_DMAABTCTX context;

  context.devCtx = devCtx;
  context.channel = chan;
  os_invoke_atomic(devCtx->irqObj, abortDMAOnHardwareCritical, &context);

  return context.success;
}

/*
** Start the next DMANODE, if it exists, on an idle DMA channel.
**
** Looks first for a DMANODE in the queue associated with the
** specified DMA channel, and if not found, looks in the 'any channel'
** queue. If a DMANODE is found, it is made active on the hardware.
**
** Must be called while holding the DMA spinlock.
*/
static void
startNextDMA(
  ADMXRC*       devCtx,
  unsigned char chan)
{
  ADMXRC_DMANODE* node;
  unsigned char   queue;

  os_dbgprint(6, ("startNextDMA: chan = %ld\n", (unsigned long) chan));
  
  /*
  ** The DMA channel is now idle - look for a candidate DMANODE
  ** to make active on the DMA channel.
  */
  node = devCtx->dma.queue[chan].head;
  if (node == NULL) {
    node = devCtx->dma.queue[devCtx->info.card.NumDMAChan].head;
    if (node != NULL) {
      /* Found one in the 'any' queue */
      queue = devCtx->info.card.NumDMAChan;
      goto found;
    }
  } else {
    /* Found one */
    queue = chan;
    goto found;
  }

  /* No DMANODE found - can't start a new DMA operation on the hardware */
  return;

 found:
  /* Found one - claim the DMA channel */
  devCtx->dma.channel[chan].active = node;
  devCtx->dma.channel[chan].activeCnt++;
  if (devCtx->dma.channel[chan].activeCnt != 1) {
    os_dbgprint(0, ("*** startNextDMA: %ld job(s) active on DMA channel %ld\n",
		    (unsigned long) devCtx->dma.channel[chan].activeCnt, (unsigned long) chan));
  }
  
  /* Dequeue the DMANODE from its queue */
  removeDMA(&devCtx->dma.queue[queue], node);
  node->system.state = 2;         /* Means active on hardware. */
  node->system.activechan = chan; /* Record which DMA channel active on */
  
  /* Kick off the DMA operation on the hardware */
  os_dbgprint(6, ("startNextDMA: Kicking off DMANODE 0x%p on DMA channel %ld\n",
		  (void*) node, (unsigned long) chan));
  doDMAChunk(devCtx, node, chan);
}

/*
** Returns TRUE if a DMANODE can be started on the hardware immediately, and
** sets the 'activechan' member of the DMA node; returns FALSE otherwise.
**
** Must be called while holding the DMA spinlock.
*/
os_boolean
ADMXRC_TestQueueDMA(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node)
{
  unsigned char queue, chan;
  os_boolean    canstart = FALSE;

  queue = node->client.queue;

  /*
  ** Attempt to bypass queueing mechanism - check if there is a DMA channel
  ** free on which to start the DMA transfer immediately
  */
  if (queue < devCtx->info.card.NumDMAChan) {
    /* Must use a specific DMA channel */
    chan = queue;
    if (devCtx->dma.channel[chan].active == NULL && devCtx->dma.queue[chan].head == NULL) {
      node->system.activechan = chan;
      canstart = TRUE;
    }
  } else {
    /* Can use any DMA channel */
    for (chan = 0; chan < devCtx->info.card.NumDMAChan; chan++) {
      if (devCtx->dma.channel[chan].active == NULL && devCtx->dma.queue[chan].head == NULL) {
	canstart = TRUE;
	node->system.activechan = chan;
	break;
      }
    }
  }

  return canstart;
}

/*
** Queue up (canstart is FALSE) a DMANODE, or starts a DMANODE on the hardware
** (canstart is TRUE).
**
** Must be called while holding the DMA spinlock.
*/
void
ADMXRC_EnqueueDMA(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node,
  os_boolean      canstart)
{
  unsigned char queue;
  unsigned char chan;

  if (canstart) {
    /* Found a free PLX DMA channel - claim it */
    chan = node->system.activechan;
    devCtx->dma.channel[chan].active = node;
    devCtx->dma.channel[chan].activeCnt++;
    if (devCtx->dma.channel[chan].activeCnt != 1) {
      os_dbgprint(0, ("*** ADMXRC_EnqueueDMA: %ld job(s) active on DMA channel %ld\n",
		      (unsigned long) devCtx->dma.channel[chan].activeCnt,
		      (unsigned long) chan));
    }
    
    node->system.state = 2;         /* Means 'active on hardware' */
    
    /* Kick off the DMA */
    os_dbgprint(6, ("ADMXRC_EnqueueDMA: Kicking off DMA for node 0x%p\n", (void*) node));
    doDMAChunk(devCtx, node, chan);
  } else {
    /*
    ** No free DMA channel on which we can start this DMA operation
    ** immediately, so add the DMANODE to the tail of the DMA queue.
    */
    queue = node->client.queue;
    os_dbgprint(6, ("ADMXRC_EnqueueDMA: queueing DMA node 0x%p in queue %ld\n",
		    (void*) node, (unsigned long) queue));
    
    addDMA(&devCtx->dma.queue[queue], node);
    
    node->system.state = 1;         /* Means 'queued' */
  }
}

/*
** This function is called to cancel a DMA transfer whose state
** is one of 'queued', 'active on the hardware', or 'completed'.
**
** If the DMA transfer was queued or active, TRUE is returned.
** If the DMA transfer was completed, FALSE is returned.
**
** After this function returns, the DMANODE is guaranteed not to
** be in any DMA queue and not active on the hardware.
*/
os_boolean
ADMXRC_CancelDMA(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node)
{
  OS_SPINLOCK_FLAGS f;
  os_boolean        success;
  unsigned char     chan;
  int               state;

  os_acquire_spinlock(&devCtx->dma.lock, &f);

  state = node->system.state;

  switch (state) {
  case 1:  /* Queued */
    {
      unsigned char queue = node->client.queue;

      removeDMA(&devCtx->dma.queue[queue], node);
      success = TRUE;
      break;
    }

  case 2:  /* Active on hardware */
    {
      chan = node->system.activechan;

      /* Abort the DMA transfer on the hardware */
      abortDMAOnHardware(devCtx, chan);
      
      /*
      ** Mark the DMA channel as idle. This will prevent the
      ** 'DMA complete' DPC from completing the DMANODE if an
      ** interrupt happens to occur when we release the DMA lock
      ** below.
      */
      devCtx->dma.channel[chan].active = NULL;
      devCtx->dma.channel[chan].activeCnt--;
      
      success = TRUE;

      /* Start the next DMA on the hardware, if there is anything in the queue */
      startNextDMA(devCtx, chan);

      break;
    }

  case 3:  /* Completed */
    {
      success = FALSE;
      break;
    }

  default: /* Should never happen! */
    {
      os_dbgprint(0, ("*** ADMXRC_CancelDMA: DMA node 0x%p has invalid state (%ld)\n",
		      (void*) node, (unsigned long) state));
      success = FALSE;
      break;
    }
  }
  
  os_release_spinlock(&devCtx->dma.lock, f);

  return success;
}

/*
** This routine is called by the DMA interrupt DPC to service a DMA interrupt
** for a particular channel.
**
** May be called outside a thread context.
*/
void
ADMXRC_DMACompleted(
 ADMXRC*       devCtx,
 unsigned char chan)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_DMANODE*   node;
  
  os_acquire_spinlock(&devCtx->dma.lock, &f);
  node = devCtx->dma.channel[chan].active;
  if (node == NULL) {
    os_release_spinlock(&devCtx->dma.lock, f);
    
    os_dbgprint(0, ("*** ADMXRC_DMAComplete: no job on channel %ld!\n", (unsigned long) chan));
  } else {
    if (node->client.length == 0) {
      /*
      ** Mark the DMA channel as idle. This will prevent the
      ** cleanup routine from attempting to complete the DMANODE
      */
      devCtx->dma.channel[chan].active = NULL;
      devCtx->dma.channel[chan].activeCnt--;
      
      /* Mark the DMANODE as finished. */
      node->system.state = 3;	
      node->status.status = OS_STATUS_SUCCESS;
      
      /* Start the next operation on the DMA channel */
      startNextDMA(devCtx, chan);
      
      os_release_spinlock(&devCtx->dma.lock, f);
      
      os_dbgprint(6, ("ADMXRC_DMACompleted: calling completion callback for DMANODE 0x%p\n",
		      (void*) node));
      node->client.callback(devCtx, node);
    } else {
      /* Do another chunk of DMA on the same node */
      os_dbgprint(6, ("ADMXRC_DMACompleted: doing another chunk of DMANODE 0x%p\n",
		      (void*) node));
      doDMAChunk(devCtx, node, chan);
      os_release_spinlock(&devCtx->dma.lock, f);
    }
  }
}

/*
** DMA cleanup function; called when card is about to be closed
*/
void
ADMXRC_CleanupDMA(
  ADMXRC*        devCtx)
{
  OS_SPINLOCK_FLAGS     f;
  ADMXRC_DMANODE*	node;
  unsigned char         i;

  if (os_dbglevel(1)) {
    for (i = 0; i <= devCtx->info.card.NumDMAChan; i++) {
      os_dbgprint(1, ("ADMXRC_CleanupDMA: queue %ld length = %ld\n",
		      (unsigned long) i, (unsigned long) devCtx->dma.queue[i].length));
    }
  }

  /* Cleanup any DMA transfers that are queued */
  for (i = 0; i <= devCtx->info.card.NumDMAChan; i++) {
    while (1) {
      os_acquire_spinlock(&devCtx->dma.lock, &f);
      node = devCtx->dma.queue[i].head;
      if (node == NULL) {
	os_release_spinlock(&devCtx->dma.lock, f);
	break;
      } else {
	os_dbgprint(3, ("ADMXRC_CleanupDMA: Cleaning up DMANODE 0x%p on DMA queue %ld\n",
			(void*) node, (unsigned long) i));
	
	/* Remove this DMANODE from the queue */
	removeDMA(&devCtx->dma.queue[i], node);
	
	/* Mark the DMANODE as finished, but with an error status */
	node->system.state = 3;
	node->status.status = OS_MAKE_STATUS(ADMXRC_CLOSED);
	
	os_release_spinlock(&devCtx->dma.lock, f);
	
	node->client.callback(devCtx, node);
      }
    }
  }
  
  /* Cleanup any DMA IRP that is currently active on the hardware */
  for (i = 0; i < devCtx->info.card.NumDMAChan; i++) {
    os_acquire_spinlock(&devCtx->dma.lock, &f);
    node = devCtx->dma.channel[i].active;
    if (node != NULL) {
      os_dbgprint(1, ("ADMXRC_CleanupDMA: Cleaning up DMANODE 0x%p active on channel %ld\n",
		      (void*) node, (unsigned long) node->system.activechan));
      
      /* Mark the DMANODE as finished, but with an error status */
      node->system.state = 3;
      node->status.status = OS_MAKE_STATUS(ADMXRC_CLOSED);
      
      /* Abort the DMA transfer on the hardware */
      abortDMAOnHardware(devCtx, i);
      
      /*
      ** Mark the DMA channel as idle. This will prevent the
      ** ADMXRC_DMACompleted routine from completing the associated IO request.
      */
      devCtx->dma.channel[i].active = NULL;
      devCtx->dma.channel[i].activeCnt--;
      
      /*
      ** Start the next operation on the DMA channel - although there
      ** should not be anything queued, because of the code above that
      ** empties the DMA queues.
      */
      startNextDMA(devCtx, i);
      
      os_release_spinlock(&devCtx->dma.lock, f);
      
      node->client.callback(devCtx, node);
    } else {
      os_release_spinlock(&devCtx->dma.lock, f);
    }
  }
  
  os_dbgprint(2, ("ADMXRC_CleanupDMA: done\n"));
}
