/*
** eeprom.c
**
** Utility functions for reading and writing the EEPROM.
*/

#include "platform.h"
#include "device.h"
#include "eeprom.h"

static void
ADMXRC_ClockEprom(
  PLX9080REGS* plx,
  os_uint32    cntrl)
{	
  /* Clock low */
  os_write_register_32endian(&plx->CNTRL, cntrl | 0x01000000U);
  /* Clock high */
  os_write_register_32endian(&plx->CNTRL, cntrl);
}


/* Send an instruction to EEPROM */
static os_boolean
ADMXRC_CommandEprom9080(
  PLX9080REGS*	plx,
  os_uint32     instr,
  int           instr_len)
{	
  int       i;
  os_uint32 cntrl;
	
  cntrl = os_read_register_32endian(&plx->CNTRL);
  cntrl |=  0x02000000U; /* Chip select TRUE                 */
  cntrl &= ~0x05000000U; /* Clear instruction and clock bits */
  
  /*
  ** Toggle EEPROM's chip select (to get it out of Shift Register Mode)
  ** and assert the chip select.
  */
  os_write_register_32endian(&plx->CNTRL, cntrl & ~0x02000000U);
  os_write_register_32endian(&plx->CNTRL, cntrl);
  
  /* Send instruction to EEPROM one bit at a time */
  for (i = 0; i < instr_len; i++) {
    /* Pre-clear the EEPROM data bit */
    cntrl &= ~0x04000000U;
    /* Set the EEPROM data bit if necessary */
    cntrl |= (instr & (0x1U << (instr_len - 1))) ? 0x04000000U : 0x00000000U;
    
    /* Clock the data bit to the EEPROM */
    os_write_register_32endian(&plx->CNTRL, cntrl);
    ADMXRC_ClockEprom(plx, cntrl);
    
    instr <<= 1; /* Align next instruction bit */
  }
  
  return TRUE;
}

/* Send an instruction to EEPROM */
static os_boolean
ADMXRC_CommandEprom9656(
  PLX9656REGS* plx,
  os_uint32    instr,
  int          instr_len)
{	
  int       i;
  os_uint32 cntrl;
	
  cntrl = os_read_register_32endian(&plx->CNTRL);
  cntrl |=  0x02000000U; /* Chip select TRUE                 */
  /* Clear instruction and clock bits and set EEDO pin to output */
  cntrl &= ~0x85000000U;

  /*
  ** Toggle EEPROM's chip select (to get it out of Shift Register Mode)
  ** and assert the chip select.
  */
  os_write_register_32endian(&plx->CNTRL, cntrl & ~0x02000000U);
  os_write_register_32endian(&plx->CNTRL, cntrl);
  
  /* Send instruction to EEPROM one bit at a time */
  for (i = 0; i < instr_len; i++) {
    /* Pre-clear the EEPROM data bit */
    cntrl &= ~0x04000000U;
    /* Set the EEPROM data bit if necessary */
    cntrl |= (instr & (0x1U << (instr_len - 1))) ? 0x04000000U : 0x00000000U;
    
    /* Clock the data bit to the EEPROM */
    os_write_register_32endian(&plx->CNTRL, cntrl);
    ADMXRC_ClockEprom((PLX9080REGS*) plx, cntrl);
    
    instr <<= 1; /* Align next instruction bit */
  }
  
  cntrl |= 0x80000000U; /* Set EEDO pin to input */
  os_write_register_32endian(&plx->CNTRL, cntrl);
  
  return TRUE;
}

os_boolean
ADMXRC_ReadEprom(
  ADMXRC*         devCtx,
  unsigned long	  add,
  unsigned long*  data)
{
  PLX9080REGS*   plx = devCtx->hw.regs.pci9080;
  os_uint32      cntrl, dataRead;
  int            i, instr_len, addr_len;
  os_uint32      addr_mask;

  switch (devCtx->info.eepromType) {
  case ADMXRC_EEPROM_9346:
    addr_mask = 0x3fU;
    addr_len = 6;
    instr_len = 9;
    break;
  case ADMXRC_EEPROM_9356:
    addr_mask = 0x7fU;
    addr_len = 8;
    instr_len = 11;
    break;
  case ADMXRC_EEPROM_9366:
    addr_mask = 0xffU;
    addr_len = 8;
    instr_len = 11;
    break;
  default:
    return FALSE;
  }
  
  if (devCtx->id.pci.subDId == PCI_SUBDEVICE_ID_ADMXRC2) {
    ADMXRC_CommandEprom9656(devCtx->hw.regs.pci9656, (EE_READ << addr_len) | (add & addr_mask), instr_len);
  } else {
    ADMXRC_CommandEprom9080(devCtx->hw.regs.pci9080, (EE_READ << addr_len) | (add & addr_mask), instr_len);
  }
		
  /* Get word from EEPROM - one bit at a time */
  cntrl = os_read_register_32endian(&plx->CNTRL);
  dataRead = 0;
  for (i = 0; i < 16; i++) {
    dataRead <<= 1;
    ADMXRC_ClockEprom(plx, cntrl);
    cntrl = os_read_register_32endian(&plx->CNTRL);
    if (cntrl & 0x08000000U) {
      dataRead |= 0x1U;
    }
  }
	
  *data = dataRead;

  return TRUE;
}

os_boolean
ADMXRC_WriteEprom(
  ADMXRC*       devCtx,
  unsigned long add,
  unsigned long data)
{
  PLX9080REGS*  plx = devCtx->hw.regs.pci9080;
  int           i, instr_len, addr_len;
  os_uint32     addr_mask;
  os_uint32     cntrl;

  switch (devCtx->info.eepromType) {
  case ADMXRC_EEPROM_9346:
    addr_mask = 0x3fU;
    addr_len = 6;
    instr_len = 9;
    break;
  case ADMXRC_EEPROM_9356:
    addr_mask = 0x7fU;
    addr_len = 8;
    instr_len = 11;
    break;
  case ADMXRC_EEPROM_9366:
    addr_mask = 0xffU;
    addr_len = 8;
    instr_len = 11;
    break;
  default:
    return FALSE;
  }

  if (devCtx->id.pci.subDId == PCI_SUBDEVICE_ID_ADMXRC2) {
    /* Enable writing to EEPROM */
    ADMXRC_CommandEprom9656(devCtx->hw.regs.pci9656, EE_WREN << (addr_len - 2), instr_len);
    /* Send write command to EEPROM */
    ADMXRC_CommandEprom9656(devCtx->hw.regs.pci9656, (EE_WRITE << addr_len) | (add & addr_mask), instr_len);
  } else {
    /* Enable writing to EEPROM */
    ADMXRC_CommandEprom9080(devCtx->hw.regs.pci9080, EE_WREN << (addr_len - 2), instr_len);
    /* Send write command to EEPROM */
    ADMXRC_CommandEprom9080(devCtx->hw.regs.pci9080, (EE_WRITE << addr_len) | (add & addr_mask), instr_len);
  }
  
  /* Pre-clear data bit */
  cntrl = os_read_register_32endian(plx->CNTRL) & ~0x04000000U;
  
  if (devCtx->id.pci.subDId == PCI_SUBDEVICE_ID_ADMXRC2) {
    /* Enable EEDO */
    cntrl &= ~0x80000000U;
  }
  
  /* Write data word to EEPROM one bit at a time */
  for (i = 0; i < 16 ; i++) {
    /* Pre-clear the EEPROM data bit */
    cntrl &= ~0x04000000U;
    /* Set the EEPROM data bit if necessary */
    cntrl |= (data & 0x8000U) ? 0x04000000U : 0x00000000U;

    /* Send the data bit */
    os_write_register_32endian(&plx->CNTRL, cntrl);
    ADMXRC_ClockEprom(plx, cntrl);

    data <<= 1;
  }
  
  if (devCtx->id.pci.subDId == PCI_SUBDEVICE_ID_ADMXRC2) {
    /* Disable EEDO */
    cntrl |= 0x80000000U;
  }

  /* Activate chip's status mechanism: deselect and reselect */
  os_write_register_32endian(&plx->CNTRL, cntrl & ~0x02000000U);
  os_write_register_32endian(&plx->CNTRL, cntrl);

  /*
  ** Poll chip status until write-cycle completes (or times out)
  */
  for (i = 0; i < 2000; i++) {
    os_uint32 val;

    val = os_read_register_32endian(plx->CNTRL);
    if (val & 0x08000000) {
      break;
    }
  }
  
  /* Disable writing to EEPROM */
  if (devCtx->id.pci.subDId == PCI_SUBDEVICE_ID_ADMXRC2) {
    ADMXRC_CommandEprom9656(devCtx->hw.regs.pci9656, EE_WDS << addr_len, instr_len);
  } else {
    ADMXRC_CommandEprom9080(devCtx->hw.regs.pci9080, EE_WDS << addr_len, instr_len);
  }
  
  return TRUE;
}					
