/*
** info.c     
**
** IOCTL handlers for returning information about a card.
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "info.h"
#include "driver_ver.h"

OS_STATUS
ADMXRC2_GetBankInfo(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC2_IOCTLS_GETBANKINFO* in;
  ADMXRC2_IOCTLS_GETBANKINFO* out;
  unsigned long               index;

  in = (ADMXRC2_IOCTLS_GETBANKINFO*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  out = (ADMXRC2_IOCTLS_GETBANKINFO*) req->param.outBuf;
  if (out == NULL || req->param.outSize != sizeof(out->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  index = in->in.index;
  if (index >= devCtx->info.card.NumRAMBank) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  os_memcpy(&out->out.info, &devCtx->info.banks[index], sizeof(ADMXRC2_BANK_INFO));

  req->status.length = sizeof(out->out);

  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_GetClockType(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_GETCLOCKTYPE* ioctl;

  ioctl = (ADMXRC_IOCTLS_GETCLOCKTYPE*) req->param.outBuf;
  if (ioctl == NULL || req->param.outSize != sizeof(ioctl->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  ioctl->out.type = (unsigned long) devCtx->info.clockType; 

  req->status.length = sizeof(ioctl->out);

  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_GetCardInfo(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req,
  os_boolean     owner)
{
  ADMXRC_IOCTLS_GETCARDINFO* ioctl;
  unsigned int               i;

  ioctl = (ADMXRC_IOCTLS_GETCARDINFO*) req->param.outBuf;
  if (ioctl == NULL || req->param.outSize != sizeof(ioctl->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  ioctl->out.info.CardID = devCtx->info.card.CardID;
  for (i = 0; i < 4; i++) {
    if (devCtx->info.card.RAMBanksFitted & (1 << i)) {
      ioctl->out.info.RAMBankFitted[i] = devCtx->info.banks[i].Size;
    } else {
      ioctl->out.info.RAMBankFitted[i] = 0;
    }
  }
  ioctl->out.info.FPGAType = devCtx->info.card.FPGAType;
  ioctl->out.info.PhysicalMemoryBase = devCtx->info.spaces[0].PhysicalBase;
  if (owner) {
    ioctl->out.info.MemoryBase = (void*) devCtx->info.spaces[0].VirtualBase;
  } else {
    ioctl->out.info.MemoryBase = NULL;
  }
  ioctl->out.info.BoardRevision = (unsigned long) devCtx->info.card.BoardRevision;
  ioctl->out.info.LogicRevision = (unsigned long) devCtx->info.card.LogicRevision;
  ioctl->out.info.SerialNum = devCtx->info.card.SerialNum;
  ioctl->out.info.Timeout = 0;

  req->status.length = sizeof(ioctl->out);

  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC2_GetCardInfo(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC2_IOCTLS_GETCARDINFO* ioctl;

  ioctl = (ADMXRC2_IOCTLS_GETCARDINFO*) req->param.outBuf;
  if (ioctl == NULL || req->param.outSize != sizeof(ioctl->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  os_memcpy(&ioctl->out, &devCtx->info.card, sizeof(ADMXRC2_CARD_INFO));
  
  req->status.length = sizeof(ioctl->out);
  
  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC2_GetSpaceInfo(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req,
  os_boolean     owner)
{
  ADMXRC2_IOCTLS_GETSPACEINFO* in;
  ADMXRC2_IOCTLS_GETSPACEINFO* out;
  unsigned long                index;

  in = (ADMXRC2_IOCTLS_GETSPACEINFO*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  out = (ADMXRC2_IOCTLS_GETSPACEINFO*) req->param.outBuf;
  if (out == NULL || req->param.outSize != sizeof(out->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  index = in->in.index;
  switch (devCtx->info.card.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
    if (index >= devCtx->info.card.NumSpace) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
    if (owner) {
      out->out.info.VirtualBase = devCtx->info.spaces[index].VirtualBase;
      out->out.info.VirtualSize = devCtx->info.spaces[index].VirtualSize;
    } else {
      out->out.info.VirtualBase = NULL;
      out->out.info.VirtualSize = 0;
    }
    out->out.info.PhysicalBase = devCtx->info.spaces[index].PhysicalBase;
    out->out.info.LocalBase = devCtx->info.spaces[index].LocalBase;
    out->out.info.LocalSize = devCtx->info.spaces[index].LocalSize;
    break;
  default:
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }

  req->status.length = sizeof(out->out);

  return OS_STATUS_SUCCESS;
}

OS_STATUS 
ADMXRC_GetVersionInfo(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_GETVERSIONINFO* ioctl;

  ioctl = (ADMXRC_IOCTLS_GETVERSIONINFO*) req->param.outBuf;
  if (ioctl == NULL || req->param.outSize != sizeof(ioctl->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  ioctl->out.major = ADMXRC_DRIVER_MAJORVER;
  ioctl->out.minor = ADMXRC_DRIVER_MINORVER;

  req->status.length = sizeof(ioctl->out);
  
  return OS_STATUS_SUCCESS;
}

