/*
** loadfpga.c     
**
** IOCTL handlers for configuring/unconfiguring the FPGA.
*/

#include "platform.h"
#include "device.h"
#include "dmacore.h"
#include "dmadesc.h"
#include "ioctl.h"
#include "loadfpga.h"

static OS_STATUS
clearFPGA(
  ADMXRC* devCtx)
{
  OS_STATUS     status;
  unsigned long timeout;
  unsigned long now;
  os_boolean    success;
  os_uint8      fcon;
  os_boolean    sig;
  unsigned long delay;

  /*
  ** Pulse PROG for 20 usec - this will cause DONE to
  ** become deasserted and INIT to (eventually) become
  ** deasserted.
  */
  os_write_register_8(&devCtx->hw.control->FCON, CPLD_FCON_PROG);
  os_delay_busy_us(20);
  os_write_register_8(&devCtx->hw.control->FCON, 0);
  
  /*
  ** Delay until DONE is deasserted.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  success = FALSE;
  while (1) {
    fcon = os_read_register_8(&devCtx->hw.control->FCON);
    if (!(fcon & CPLD_FCON_DONE)) {
      success = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      /* Timeout occurred */
      break;
    }

    delay = os_ticks_per_second / 100;
    sig = os_delay_thread(&delay);
    if (sig) {
      /* Calling thread is being terminated? */
      os_dbgprint(3, ("clearFPGA: signal while waiting for DONE deasserted\n"));
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      goto done;
    }
  }
  
  if (!success) {
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
    goto done;
  }
  
  /*
  ** Delay until INIT is deasserted.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  success = FALSE;
  while (1) {
    fcon = os_read_register_8(&devCtx->hw.control->FCON);
    if (!(fcon & CPLD_FCON_INIT)) {
      success = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      /* Timeout occurred */
      break;
    }

    delay = os_ticks_per_second / 100;
    sig = os_delay_thread(&delay);
    if (sig) {
      /* Calling thread is being terminated? */
      os_dbgprint(3, ("clearFPGA: signal while waiting for INIT deasserted\n"));
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      goto done;
    }
  }
  
  if (!success) {
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
    goto done;
  }
  
  status = OS_STATUS_SUCCESS;
  
 done:
  return status;
}

static OS_STATUS
checkConfigured(
  ADMXRC* devCtx)
{
  OS_STATUS     status;
  os_boolean    success;
  unsigned long timeout;
  unsigned long now;
  os_uint8      fcon;
  os_boolean    sig;
  unsigned long delay;

  /*
  ** Wait for device to assert DONE, which may rely on DLL's/DCM's.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  success = FALSE;
  while (1)	{
    fcon = os_read_register_8(&devCtx->hw.control->FCON);
    if (fcon & CPLD_FCON_DONE) {
      success = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      /* Timeout occurred */
      break;
    }

    delay = os_ticks_per_second / 100;
    sig = os_delay_thread(&delay);
    if (sig) {
      status = OS_MAKE_STATUS(ADMXRC_CANCELLED);
      goto done;
    }
  }
  
  if (!success) {
    if (!(os_read_register_8(&devCtx->hw.control->FCON) & CPLD_FCON_DONE)) {
      if (os_read_register_8(&devCtx->hw.control->FCON) & CPLD_FCON_INIT) {
	os_dbgprint(3, ("checkConfigured: Init indicates failure to load\n"));
      } else {
	os_dbgprint(3, ("checkConfigured: Done did not go high\n"));
      }
    }
    
    /*
    ** Something bad happened, so clear the FPGA in case
    ** it has been loaded with a bad bitstream.
    */
    os_write_register_8(&devCtx->hw.control->FCON, CPLD_FCON_PROG);
    os_delay_busy_us(20);
    os_write_register_8(&devCtx->hw.control->FCON, 0);
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
  }
  else {
    status = OS_STATUS_SUCCESS;
  }
  
 done:
  return status;
}

/*
** IOCTL handlers
*/

OS_STATUS
ADMXRC_LoadFPGA(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  os_uint8*               data;
  unsigned long           length;
  unsigned long           count;
  OS_STATUS               status;
  OS_SPINLOCK_FLAGS       f;

  os_dbgprint(3, ("ADMXRC_LoadFPGA: entered\n"));

  data = (os_uint8*) req->param.massBuf;
  length = req->param.massSize;

  os_dbgprint(3, ("ADMXRC_LoadFPGA: data = 0x%p length = %ld\n", (void*) data, length));

  os_acquire_spinlock(&devCtx->fpga.lock, &f);
  if (devCtx->fpga.busy) {
    os_release_spinlock(&devCtx->fpga.lock, f);
    return OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
  } else {
    devCtx->fpga.busy = TRUE;
    os_release_spinlock(&devCtx->fpga.lock, f);
  }
  
  status = clearFPGA(devCtx);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGA: failed to clear FPGA\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGA: cleared FPGA\n"));

  /* Send the bitstream to the FPGA */
  count = length;
  for (count = 0; count < length; count++) {
    os_write_register_8(&devCtx->hw.control->SELECTMAP, *data++);
  }
  
  status = checkConfigured(devCtx);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGA: configuration failed\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGA: success\n"));

  status = OS_STATUS_SUCCESS;
  
 done:
  os_acquire_spinlock(&devCtx->fpga.lock, &f);
  devCtx->fpga.busy = FALSE;
  os_release_spinlock(&devCtx->fpga.lock, f);
  
  return status;
}

static void
loadFPGADMACallback(
  ADMXRC*         devCtx,
  ADMXRC_DMANODE* node)
{
  ADMXRC_LOADFPGADMA* req;
  
  os_dbgprint(6, ("loadFPGADMACallback\n"));
  
  req = (ADMXRC_LOADFPGADMA*) node->client.context;
  
  os_signal_event(&req->event);
}

OS_STATUS
ADMXRC_LoadFPGADMA(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_LOADFPGADMA* iptr;
  void*                      buffer;
  unsigned long              length;
  OS_STATUS                  status;
  OS_SPINLOCK_FLAGS          f;
  ADMXRC_LOADFPGADMA*        ctl = NULL;
  ADMXRC_DMANODE*            dmanode;
  ADMXRC_DMATGT*             dmatgt = NULL;
  os_uint32                  local;
  unsigned long              flags;
  unsigned char              channel, queue;
  os_boolean                 canstart, signalled;

  os_dbgprint(0, ("ADMXRC_LoadFPGADMA: entered\n"));

  iptr = (ADMXRC_IOCTLS_LOADFPGADMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(*iptr)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  flags = iptr->in.flags;
  if (flags & ~(ADMXRC_DMAFLAG_TIMEOUT | ADMXRC_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  channel = iptr->in.channel;
  if (channel != ADMXRC_DMACHAN_ANY && channel >= devCtx->info.card.NumDMAChan) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  length = iptr->in.length;
  buffer = iptr->in.buffer;
  if (buffer == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NULL_POINTER);
  }

  ctl = os_alloc_lcache(devCtx->fpga.ctlPool, 0);
  if (ctl == NULL) {
    return OS_MAKE_STATUS(ADMXRC_NO_MEMORY);
  }

  status = ADMXRC_CreateDMATarget(devCtx, buffer, length, &dmatgt);
  if (!OS_SUCCESS(status)) {
    goto done;
  }

  os_acquire_spinlock(&devCtx->fpga.lock, &f);
  if (devCtx->fpga.busy) {
    os_release_spinlock(&devCtx->fpga.lock, f);
    status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
    goto done;
  } else {
    devCtx->fpga.busy = TRUE;
    os_release_spinlock(&devCtx->fpga.lock, f);
  }

  status = clearFPGA(devCtx);
  if (!OS_SUCCESS(status)) {
    goto done;
  }

  /* Determine which DMA queue to add this node to */
  queue = (channel == ADMXRC_DMACHAN_ANY) ? devCtx->info.card.NumDMAChan : (unsigned char) channel;
  
  /* Calculate local bus address of SelectMap port */
  local = devCtx->info.spaces[1].LocalBase + 0x8;
  
  os_init_event(&ctl->event);
  dmanode = &ctl->dmanode;
  dmanode->client.context = ctl;
  dmanode->client.target = dmatgt;
  dmanode->client.callback = loadFPGADMACallback;
  dmanode->client.direction = (unsigned char) ADMXRC2_PCITOLOCAL;
  dmanode->client.length = length;
  dmanode->client.local = local;
  dmanode->client.pos = os_get_usbd_page_offset(dmatgt->usbd);
  dmanode->client.mode = 0x00020dc0U | (ADMXRC_ALLOW_WRITE_INVALIDATE << 13U);
  dmanode->client.queue = queue;
  
  /* Queue up the DMA (it might begin on hardware immediately) */
  os_acquire_spinlock(&devCtx->dma.lock, &f);
  canstart = ADMXRC_TestQueueDMA(devCtx, dmanode);
  if ((flags & ADMXRC_DMAFLAG_DONOTQUEUE) && !canstart) {
    os_release_spinlock(&devCtx->dma.lock, f);
    status = OS_MAKE_STATUS(ADMXRC_DEVICE_BUSY);
    goto done;
  } else {
    ADMXRC_EnqueueDMA(devCtx, dmanode, canstart);
    os_release_spinlock(&devCtx->dma.lock, f);
    status = OS_STATUS_PENDING;
  }
  
  /*
  ** Wait for DMA transfer to either finish/be cancelled/fail, OR
  ** for a CTRL-C (ie. user aborts the program)
  */
  signalled = os_wait_event(&ctl->event, TRUE);
  if (signalled) {
    ADMXRC_CancelDMA(devCtx, dmanode);
    status = ADMXRC_CANCELLED;
    goto done;
  }

  status = ctl->dmanode.status.status;
  if (!OS_SUCCESS(status)) {
    goto done;
  }

  status = checkConfigured(devCtx);
  
 done:
  os_free_lcache(devCtx->fpga.ctlPool, ctl);

  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }

  os_acquire_spinlock(&devCtx->fpga.lock, &f);
  devCtx->fpga.busy = FALSE;
  os_release_spinlock(&devCtx->fpga.lock, f);
  
  return status;
}

