/*
** main.c
**
** Generic entry points for the ADM-XRC series driver, called by the
** framework.
*/

#include "platform.h"
#include "device.h"
#include "device_defs.h"
#include "ioctl.h"

/* Helper routines for getting info about card */
#include "cardinfo.h"

/* The following header files declare the IOCTL handlers */
#include "configrw.h"
#include "dma.h"
#include "dmacore.h"
#include "dmadesc.h"
#include "info.h"
#include "intwait.h"
#include "loadfpga.h"
#include "owner.h"
#include "rw.h"
#include "setclock.h"

static OS_DEVICE_DESC devDescs[4];

/*
** Declare forward references
*/

static OS_STATUS
onStopDevice(
  OS_DEVICE_OBJECT*    devObj);

static void
freeResources(
  ADMXRC* devCtx);

static void
dmaint_dpc(
  void* context,
  void* argument);

/*
** -----------------------------------------------------------------------------
** Helper routines
** -----------------------------------------------------------------------------
*/

static OS_STATUS
allocateResources(
  ADMXRC* devCtx)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  int       i;

  os_dbgprint(1, ("allocateResources: entered\n"));

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    unsigned long n = sizeof(PLXDESC) * (ADMXRC_MAX_PLXDESC_PER_CHANNEL + 1);
    int           j;

    /* Allocate a block of PLX descriptors for the DMA channel, with one spare for alignment purposes */
    devCtx->dma.channel[i].descriptors.real = os_kernel_malloc_dma(n);
    if (devCtx->dma.channel[i].descriptors.real == NULL) {
      os_dbgprint(0, ("allocateResources: failed to allocate descriptors for DMA channel %d (requested %ld bytes)\n",
		      i, n));
      goto failed;
    }

    /* Align the PLX descriptors to a 16-byte boundary */
    devCtx->dma.channel[i].descriptors.aligned =
      (PLXDESC*) ((((unsigned long) devCtx->dma.channel[i].descriptors.real) + sizeof(PLXDESC) - 1) & ~(sizeof(PLXDESC) - 1));

    /* Make a lookup table of the bus-relative addresses of the PLX descriptors */
    for (j = 0; j < ADMXRC_MAX_PLXDESC_PER_CHANNEL; j++) {
      devCtx->dma.channel[i].descriptors.pci[j] = 
	(os_uint32) os_kernel_to_bus(devCtx->devObj, &devCtx->dma.channel[i].descriptors.aligned[j]);
    }
  }

  status = os_create_dpc(devCtx->devObj, ADMXRC_FpgaIntDpc, devCtx, &devCtx->interrupts.fpgaintDpc);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(0, ("allocateResources: failed to create DPC for FPGA interrupt\n"));
    goto failed;
  }

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    status = os_create_dpc(devCtx->devObj, dmaint_dpc, devCtx, &devCtx->interrupts.dmaintDpc[i]);
    if (!OS_SUCCESS(status)) {
      os_dbgprint(0, ("allocateResources: failed to create DPC for DMA interrupt\n"));
      goto failed;
    }
  }

  devCtx->dmadesc.table = (ADMXRC_DMATGT**) os_kernel_malloc(sizeof(ADMXRC_DMATGT*) * ADMXRC_MAX_DMADESC);
  if (devCtx->dmadesc.table == NULL) {
    os_dbgprint(0, ("allocateResources: failed to allocate DMA target mapping table\n"));
    goto failed;
  }
  for (i = 0; i < ADMXRC_MAX_DMADESC; i++) {
    devCtx->dmadesc.table[i] = NULL;
  }

  devCtx->dmadesc.freestack = (int*) os_kernel_malloc(sizeof(int) * ADMXRC_MAX_DMADESC);
  if (devCtx->dmadesc.freestack == NULL) {
    os_dbgprint(0, ("allocateResources: failed to allocate DMA descriptor free stack\n"));
    goto failed;
  }
  for (i = 0; i < ADMXRC_MAX_DMADESC; i++) {
    devCtx->dmadesc.freestack[i] = ADMXRC_MAX_DMADESC - 1 - i;
  }

  status = os_create_lcache(sizeof(ADMXRC_DMANODE), 0, &devCtx->dma.ctlPool);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  status = os_create_lcache(sizeof(ADMXRC_LOADFPGADMA), 0, &devCtx->fpga.ctlPool);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  return status;

 failed:
  freeResources(devCtx);
  return status;
}

static void
freeResources(
  ADMXRC* devCtx)
{
  int i;

  os_dbgprint(1, ("freeResources: entered\n"));

  if (devCtx->fpga.ctlPool != NULL) {
    os_destroy_lcache(devCtx->fpga.ctlPool);
    devCtx->fpga.ctlPool = NULL;
  }

  if (devCtx->dma.ctlPool != NULL) {
    os_destroy_lcache(devCtx->dma.ctlPool);
    devCtx->dma.ctlPool = NULL;
  }

  if (devCtx->dmadesc.freestack != NULL) {
    os_kernel_free(devCtx->dmadesc.freestack);
    devCtx->dmadesc.freestack = NULL;
  }

  if (devCtx->dmadesc.table == NULL) {
    os_kernel_free(devCtx->dmadesc.table);
    devCtx->dmadesc.table = NULL;
  }

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    if (devCtx->interrupts.dmaintDpc[i] != NULL) {
      os_destroy_dpc(devCtx->interrupts.dmaintDpc[i]);
      devCtx->interrupts.dmaintDpc[i] = NULL;
    }
  }

  if (devCtx->interrupts.fpgaintDpc != NULL) {
    os_destroy_dpc(devCtx->interrupts.fpgaintDpc);
    devCtx->interrupts.fpgaintDpc = NULL;
  }

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    if (devCtx->dma.channel[i].descriptors.real != NULL) {
      os_kernel_free_dma(devCtx->dma.channel[i].descriptors.real);
      devCtx->dma.channel[i].descriptors.real = NULL;
    }
  }
}

static void*
enableInterrupts(
  void* context)
{
  ADMXRC*   devCtx = (ADMXRC*) context;  
  os_uint32 intcsr;
	
  os_dbgprint(1, ("enableInterrupts: entered\n"));

  /* Establish the PLX DMACSR register: both DMA channels initially enabled */
  devCtx->shadow.dmacsr = 0x0101U;
  os_write_register_16endian((os_uint16*) &devCtx->hw.regs.pci9080->DMACSR, devCtx->shadow.dmacsr);

  /* Clear any outstanding FPGA interrupts */
  os_write_register_8(&devCtx->hw.control->ICON, 1); 
	
  /* Unmask the FPGA interrupt */
  os_write_register_8(&devCtx->hw.control->IMCLR, 1); 
	
  /* Enable PLX interrupts */
  os_write_register_32endian(&devCtx->hw.regs.pci9080->INTCSR,
			     PLX_INTCSR_PCI_IE | PLX_INTCSR_LOC_IE | PLX_INTCSR_DMA0_IE | PLX_INTCSR_DMA1_IE);

  intcsr = os_read_register_32endian(&devCtx->hw.regs.pci9080->INTCSR);
  os_dbgprint(1, ("enableInterrupts: interrupts enabled, PLX.INTCSR = 0x%8.8lx\n",
		  (unsigned long) intcsr));
  
  devCtx->interrupts.enabled = TRUE;

  return (void*) OS_STATUS_SUCCESS;
}

static void*
disableInterrupts(
  void* context)
{
  ADMXRC*   devCtx = (ADMXRC*) context;  
  os_uint32 intcsr;
	
  os_dbgprint(1, ("disableInterrupts: entered\n"));

  if (devCtx->interrupts.enabled) {
    /* Disable any interrupts from the PLX */
    os_write_register_32endian(&devCtx->hw.regs.pci9080->INTCSR, 0);
    
    /* Mask the FPGA interrupt */
    os_write_register_8(&devCtx->hw.control->IMSET, 1);

    intcsr = os_read_register_32endian(&devCtx->hw.regs.pci9080->INTCSR);
    os_dbgprint(1, ("disableInterrupts: interrupts disabled, PLX.INTCSR = 0x%8.8lx\n",
		    (unsigned long) intcsr));
  
    devCtx->interrupts.enabled = FALSE;
  } else {
    os_dbgprint(1, ("disableInterrupts: no need to disable interrupts\n"));
  }

  return (void*) OS_STATUS_SUCCESS;
}

static void
initCard9080(
  ADMXRC* devCtx)
{
  os_dbgprint(1, ("initCard9080: entered\n"));

  /* Enable Direct Master access */
  os_write_register_32endian(&devCtx->hw.regs.pci9080->DMRR,   0xFF000000U);
  os_write_register_32endian(&devCtx->hw.regs.pci9080->DMLBAM, 0xFF000000U);
  os_write_register_32endian(&devCtx->hw.regs.pci9080->DMLBAI, 0x0U);
  os_write_register_32endian(&devCtx->hw.regs.pci9080->DMPBAM, 0x1U);
}

static void
initCard9656(
  ADMXRC* devCtx)
{
  /*
  ** Currently, as far as this driver is concerned, card initialization
  ** for the PCI9656 is the same as for the PCI9080, but with a few extra
  ** bits and pieces.
  */
  initCard9080(devCtx);

  if (ADMXRC_9656_ERRATA_WORKAROUND) {
    os_uint32 val;

    /*
    ** Implement workarounds for PCI9656 rev AC and AD errata
    */
    
    /* DMPBAM register */
    val = os_read_register_32endian(&devCtx->hw.regs.pci9656->DMPBAM);
    val &= ~(0x1U <<  2); /* Disable Direct Master read-ahead */
    /* Do not prefetch more than 8 QWORDs for a Direct Master read */
    val &= ~(0x1U <<  3);
    val |=  (0x1U << 12);
    os_write_register_32endian(&devCtx->hw.regs.pci9656->DMPBAM, val);
    
    /* EROMBA register */
    val = os_read_register_32endian(&devCtx->hw.regs.pci9656->EROMBA);
    val &= ~(0x1U <<  4); /* Don't use local bus backoff for Direct Master transfers */
    os_write_register_32endian(&devCtx->hw.regs.pci9656->EROMBA, val);
    
    /* LBRD0 register */
    val = os_read_register_32endian(&devCtx->hw.regs.pci9656->LBRD0);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 0 reads */
    os_write_register_32endian(&devCtx->hw.regs.pci9656->LBRD0, val);
    
    /* LBRD1 register */
    val = os_read_register_32endian(&devCtx->hw.regs.pci9656->LBRD1);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 1 reads */
    os_write_register_32endian(&devCtx->hw.regs.pci9656->LBRD1, val);
    
    /* LMISC2 register */
    val = os_read_register_8endian(&devCtx->hw.regs.pci9656->LMISC2);
    val |=  (0x1U <<  5); /* Reserve 1 location in Direct Slave write FIFO */
    os_write_register_8endian(&devCtx->hw.regs.pci9656->LMISC2, val);
    
    /* MARBR register */
    val = os_read_register_32endian(&devCtx->hw.regs.pci9656->MARBR);
    val &= ~(0x1U << 16); /* Disable local bus latency timer */
    val &= ~(0x1U << 23); /* Assert REQ# on PCI for entire DMA transfer */
    val |=  (0x1U << 24); /* Operate in PCI 2.1 mode */
    val |=  (0x1U << 26); /* Flush Direct Slave read FIFO after Direct Slave write */
    val |=  (0x1U << 28); /* Direct slave read ahead mode enabled */
    os_write_register_32endian(&devCtx->hw.regs.pci9656->MARBR, val);
  }
}

static void
initCard(
  ADMXRC* devCtx)
{
  os_dbgprint(1, ("initCard: entered\n"));
  
  switch (devCtx->id.pci.subDId) {
  case PCI_SUBDEVICE_ID_ADMXRC:
  case PCI_SUBDEVICE_ID_ADMXRC2_L:
    initCard9080(devCtx);
    break;
  case PCI_SUBDEVICE_ID_ADMXRC2:
    initCard9656(devCtx);
    break;
  default:
    os_dbgprint(0, ("initCard: unrecognised hardware ID: 0x%4.4lx%4.4lx 0x%4.4lx%4.4lx\n",
		    (unsigned long) devCtx->id.pci.vId,
		    (unsigned long) devCtx->id.pci.dId,
		    (unsigned long) devCtx->id.pci.subVId,
		    (unsigned long) devCtx->id.pci.subDId));
    break;
  }
}

/*
** -----------------------------------------------------------------------------
** Interrupt handlers
** -----------------------------------------------------------------------------
*/

static void
dmaint_dpc(
  void* context,
  void* argument)
{
  ADMXRC*       devCtx = (ADMXRC*) context;
  unsigned char chan = (unsigned char) (int) argument;

  ADMXRC_DMACompleted(devCtx, chan);
}

static os_boolean
onInterrupt(
  void* context)
{
  ADMXRC*    devCtx = (ADMXRC*) context;
  os_uint32  intcsr;
  os_uint8   icon;
  os_uint16  dmacsr;
  int        dma_pending;
  os_boolean handled = FALSE;

  if (!devCtx->interrupts.enabled) {
    return handled;
  }

  intcsr = os_read_register_32endian(&devCtx->hw.regs.pci9080->INTCSR);
	
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (intcsr & PLX_INTCSR_LOC_INT) {
    icon = os_read_register_8(&devCtx->hw.control->ICON);

    if (icon) {
      os_dbgprint(6, ("onInterrupt: FPGA interrupt\n"));

      /* Clear the FPGA interrupt in the CPLD */
      os_write_register_8(&devCtx->hw.control->ICON, icon);

      /* Schedule the DPC that does the real work of handling the FPGA interrupt */
      os_schedule_dpc(devCtx->interrupts.fpgaintDpc, NULL);
    }
    handled = TRUE;
  }

  /* Check for abort conditions */
  if (intcsr & PLX_INTCSR_PCI_ABT_INT) {
    /*
    ** One or more abort conditions is present.
    **
    ** Find out exactly which, and take appropriate action to clear it.
    */
    if (intcsr & PLX_INTCSR_DM_ABT_INT) {
      /* Abort while doing Direct Master transfer */
      os_dbgprint(0, ("*** onInterrupt: Direct Master abort interrupt\n"));
    }

    if (intcsr & PLX_INTCSR_DMA0_ABT_INT) {
      /* Abort while doing Channel 0 DMA */
      os_dbgprint(0, ("*** onInterrupt: DMA channel 0 abort interrupt\n"));
    }
    
    if (intcsr & PLX_INTCSR_DMA1_ABT_INT) {
      /* Abort while doing Channel 1 DMA */
      os_dbgprint(0, ("*** onInterrupt: DMA channel 1 abort interrupt\n"));
    }

    os_dbgprint(0, ("                 Abort type: %s abort\n",
		    (intcsr & PLX_INTCSR_TGT_ABT_INT) ? "master" : "target"));
  }

  /* Is there PCI9080/PCI9656 DMA interrupt? */
  if (intcsr & (PLX_INTCSR_DMA0_INT | PLX_INTCSR_DMA1_INT)) {
    os_dbgprint(6, ("onInterrupt: DMA interrupt\n"));
    
    /* Find out which DMA channels are interrupting */
    dmacsr = 0x0U;
    dma_pending = 0x0U;
    if (intcsr & (PLX_INTCSR_DMA0_INT)) {
      dmacsr |= 0x0008U;
      dma_pending |= 0x1U;
    }
    if (intcsr & (PLX_INTCSR_DMA1_INT)) {
      dmacsr |= 0x0800U;
      dma_pending |= 0x2U;
    }
    
    if (dma_pending) {
      /* Clear the pending DMA interrupts(s) */
      os_write_register_16endian((os_uint16*) &devCtx->hw.regs.pci9080->DMACSR,
				 (os_uint16) (dmacsr | devCtx->shadow.dmacsr));  

      /* Schedule the DPC that does the real work of handling the DMA interrupt */
      if (dma_pending & 0x1U) {
	os_schedule_dpc(devCtx->interrupts.dmaintDpc[0], (void*) 0);
      }
      if (dma_pending & 0x2U) {
	os_schedule_dpc(devCtx->interrupts.dmaintDpc[1], (void*) 1);
      }
    }
    
    handled = TRUE;
  }

  return handled;
}

/*
** -----------------------------------------------------------------------------
** OS_DRIVER_OBJECT methods
** -----------------------------------------------------------------------------
*/

static OS_STATUS
onCreateDevice(
  OS_DRIVER_OBJECT*    drvObj,
  OS_DEVICE_OBJECT*    devObj,
  int                  matchIndex,
  OS_DEVICE_DESC*      id)
{
  ADMXRC*   devCtx = NULL;
  OS_STATUS status = OS_STATUS_SUCCESS;
  int       i;

  os_dbgprint(1, ("onCreateDevice: entered\n"));

  switch (id->type) {
  case OS_DEVICE_DESC_TYPE_PCI:
    os_dbgprint(3, ("PCI device: id = %4.4x:%4.4x subid = %4.4x:%4.4x\n",
		 id->pci.vId, id->pci.dId, id->pci.subVId, id->pci.subDId));
    break;
  default:
    os_dbgprint(0, ("onCreateDevice: don't know how to handle device of type %d\n", id->type));
    status = OS_STATUS_CONFIGERROR;
    goto failed;
  }
  
  devCtx = (ADMXRC*) os_kernel_malloc(sizeof(ADMXRC));
  if (devCtx == NULL) {
    os_dbgprint(0, ("onCreateDevice: failed to allocate memory for device context\n"));
    status = OS_STATUS_NOMEMORY;
    goto failed;
  }

  /* Important to zero the entire device context to avoid uninitialised variables */
  memset(devCtx, 0, sizeof(ADMXRC));

  devCtx->drvObj = drvObj;
  devCtx->devObj = devObj;
  memcpy(&devCtx->id, id, sizeof(OS_DEVICE_DESC));

  /* Interrupts are initially disabled */
  devCtx->interrupts.enabled = FALSE;
  
  /*
  ** Init FPGA stuff
  */
  devCtx->fpga.ctlPool = NULL;
  devCtx->fpga.busy = FALSE;
  os_init_spinlock(&devCtx->fpga.lock);

  /*
  ** Init interrupt wait stuff
  */
  devCtx->intwait.signalled = FALSE;
  devCtx->intwait.waiting.req = NULL;
  os_init_spinlock(&devCtx->intwait.lock);

  /*
  ** Init owner
  */
  devCtx->owner.handle = NULL;
  os_init_spinlock(&devCtx->owner.lock);

  /*
  ** Init DMA descriptor stuff
  */
  devCtx->dmadesc.table = NULL;
  devCtx->dmadesc.freestack = NULL;
  devCtx->dmadesc.top = ADMXRC_MAX_DMADESC;
  os_init_spinlock(&devCtx->dmadesc.lock);

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    devCtx->dma.channel[i].descriptors.real = NULL;
    devCtx->dma.channel[i].active = NULL;
    devCtx->dma.channel[i].activeCnt = 0;
  }
  os_init_spinlock(&devCtx->dma.lock);
  devCtx->dma.ctlPool = NULL;
  for (i = 0; i <= ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    devCtx->dma.queue[i].head = devCtx->dma.queue[i].tail = NULL;
    devCtx->dma.queue[i].length = 0;
  }

  status = allocateResources(devCtx);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  os_set_device_context(devObj, devCtx);

  return status;

 failed:
  if (devCtx != NULL) {
   freeResources(devCtx);
   os_kernel_free(devCtx);
  }
  return status;
}

static OS_STATUS
onStartDevice(
  OS_DEVICE_OBJECT*    devObj,
  OS_DEVICE_RESOURCES* resource)
{
  const unsigned long lengths[4] = { 0x100, 0x0, 0x400000, 0x400000 };
  const os_boolean    map[4] = { TRUE, FALSE, TRUE, TRUE };
  ADMXRC*             devCtx;
  int                 i;
  OS_STATUS           status = OS_STATUS_SUCCESS;

  os_dbgprint(1, ("onStartDevice: entered\n"));

  devCtx = os_get_device_context(devObj);

  for (i = 0; i < resource->raw.pci.count; i++) {
    os_dbgprint(3, ("            BAR%d io = %d physbase = 0x%8.8lx%8.8lx physlen = 0x%8.8lx0x%8.8lx\n",
		    i,
		    resource->raw.pci.io[i],
		    (unsigned long) os_uint64_hi(resource->raw.pci.base[i]),
		    (unsigned long) os_uint64_lo(resource->raw.pci.base[i]),
		    (unsigned long) os_uint64_hi(resource->raw.pci.size[i]),
		    (unsigned long) os_uint64_lo(resource->raw.pci.size[i])));
  }
  os_dbgprint(3, ("            IRQ phys = %ld\n", (unsigned long) resource->raw.pci.irq_line));
  for (i = 0; i < resource->translated.pci.count; i++) {
    os_dbgprint(3, ("            BAR%d io = %d sysbase  = 0x%8.8lx%8.8lx syslen  = 0x%8.8lx%8.8lx\n",
		    i,
		    resource->translated.pci.io[i],
		    (unsigned long) os_uint64_hi(resource->translated.pci.base[i]),
		    (unsigned long) os_uint64_lo(resource->translated.pci.base[i]),
		    (unsigned long) os_uint64_hi(resource->translated.pci.size[i]),
		    (unsigned long) os_uint64_lo(resource->translated.pci.size[i])));
  }
  os_dbgprint(3, ("            IRQ sys  = %ld\n", resource->translated.pci.vector));

  /* Save a copy of the resources */
  memcpy(&devCtx->resources, resource, sizeof(OS_DEVICE_RESOURCES));

  /*
  ** Check that the BARs have at least the minimum sizes
  */
  if (devCtx->resources.translated.pci.count < 4) {
    os_dbgprint(0, ("onStartDevice: fewer than 4 BARs\n"));
    status = OS_STATUS_CONFIGERROR;
    goto failed;
  }
  for (i = 0; i < 4; i++) {
    if (devCtx->resources.translated.pci.size[i] < lengths[i]) {
      os_dbgprint(0, ("onStartDevice: BAR%d is less than 0x%8.8lx%8.8lx bytes long\n",
		      i,
		      (unsigned long) os_uint64_hi(devCtx->resources.translated.pci.size[i]),
		      (unsigned long) os_uint64_lo(devCtx->resources.translated.pci.size[i])));
      status = OS_STATUS_CONFIGERROR;
      goto failed;
    }
  }

  /*
  ** Map the BARs into kernel address space
  */
  for (i = 0; i < 4; i++) {
    if (map[i]) {
      os_dbgprint(1, ("onStartDevice: mapping BAR%d\n", i));
      devCtx->bar[i] = os_map_kernel(devCtx->resources.translated.pci.base[i],
				     devCtx->resources.translated.pci.size[i]);
      if (devCtx->bar[i] == NULL) {
	os_dbgprint(0, ("onStartDevice: failed to map BAR%d\n", i));
	status = OS_STATUS_CONFIGERROR;
	goto failed;
      }
    } else {
      devCtx->bar[i] = NULL;
    }
  }

  /*
  ** Attach the interrupt handler
  */
  os_dbgprint(1, ("onStartDevice: attaching interrupt handler\n"));
  status = os_attach_interrupt(devObj,
			       onInterrupt,
			       devCtx->resources.translated.pci.vector,
			       DRIVER_NAME,
			       &devCtx->irqObj);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(0, ("onStartDevice: failed to attach interrupt handler\n"));
    goto failed;
  }

  devCtx->hw.regs.untyped = (void*) devCtx->bar[0];
  devCtx->hw.fpga = (void*) devCtx->bar[2];
  devCtx->hw.control = (CONTROLXRC*) devCtx->bar[3];
  devCtx->hw.flash = (void*) (((os_uint8*) devCtx->bar[3]) + 0x200000U);

  /*
  ** Initialise the registers on the card.
  */
  initCard(devCtx);

  /*
  ** Obtain the physical configuration of the card, ie. how much memory etc.
  ** by reading the EEPROM (or otherwise)
  */
  if (!ADMXRC_SetupCardInfo(devCtx)) {
    os_dbgprint(0, ("onStartDevice: failed to read card information from EEPROM\n"))
    goto failed;
  }

  /*
  ** Declare the regions that are mappable to user space
  **
  ** space 0 => BAR 2
  ** space 1 => BAR 3
  */
  devCtx->userMapped.count = 0;
  switch (devCtx->info.card.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
    {
      for (i = 0; i < devCtx->info.card.NumSpace; i++) {
	if (devCtx->resources.translated.pci.size[i + 2]) {
	  status = os_allow_mapping(devObj,
				    devCtx->resources.translated.pci.base[i + 2],
				    devCtx->resources.translated.pci.size[i + 2],
				    (void*) i, /* context value */
				    &devCtx->userMapped.regions[i]);
	  devCtx->userMapped.count++;
	}
      }
    }
    break;
  default:
    break;
  }

  /* Enable interrupts - done with interrupts disabled to avoid races */
  os_invoke_atomic(devCtx->irqObj, enableInterrupts, devCtx);

  return status;

 failed:
  onStopDevice(devObj);
  return status;
}

static OS_STATUS
onStopDevice(
  OS_DEVICE_OBJECT*    devObj)
{
  ADMXRC*   devCtx;
  int       i;
  OS_STATUS status = OS_STATUS_SUCCESS;

  os_dbgprint(1, ("onStopDevice: entered\n"));

  devCtx = os_get_device_context(devObj);

  /* Disable interrupts - done with interrupts disabled to avoid races */
  os_invoke_atomic(devCtx->irqObj, disableInterrupts, devCtx);

  /*
  ** Unmap regions from user space
  */
  for (i = 0; i < devCtx->userMapped.count; i++) {
    if (devCtx->userMapped.regions[i] != NULL) {
      os_disallow_mapping(devObj,
			  devCtx->userMapped.regions[i]);
      devCtx->userMapped.regions[i] = NULL;
    } else {
      os_dbgprint(0, ("onStopDevice: region %d is NULL\n", i));
    }
  }

  /*
  ** Detach the interrupt handler
  */
  os_dbgprint(1, ("onStopDevice: detaching interrupt handler\n"));
  if (devCtx->irqObj != NULL) {
    os_detach_interrupt(devCtx->irqObj); 
    devCtx->irqObj = NULL;
  }

  /*
  ** Unmap the BARs from kernel address space
  */
  for (i = 0; i < 4; i++) {
    if (devCtx->bar[i] != NULL) {
      os_dbgprint(1, ("onStopDevice: unmapping BAR%d\n", i));
      os_unmap_kernel(devCtx->bar[i],
		      devCtx->resources.translated.pci.base[i],
		      devCtx->resources.translated.pci.size[i]);
      devCtx->bar[i] = NULL;
    }
  }

  return status;
}

static void
onDeleteDevice(
  OS_DEVICE_OBJECT*    devObj)
{
  ADMXRC* devCtx;

  os_dbgprint(1, ("onDeleteDevice: entered\n"));

  devCtx = os_get_device_context(devObj);

  freeResources(devCtx);

  os_kernel_free(devCtx);
}

static OS_STATUS
onOpen(
  OS_DEVICE_OBJECT* devObj,
  OS_USER_CLIENT    client)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  ADMXRC*   devCtx;

  os_dbgprint(1, ("onOpen: entered, client = 0x%p\n", client));

  devCtx = os_get_device_context(devObj);

  return status;
}

static OS_STATUS
onClose(
  OS_DEVICE_OBJECT* devObj,
  OS_USER_CLIENT    client)
{
  OS_STATUS  status = OS_STATUS_SUCCESS;
  ADMXRC*    devCtx;

  os_dbgprint(1, ("onClose: entered, client = 0x%p\n", client));

  devCtx = os_get_device_context(devObj);

  if (client == devCtx->owner.handle) {
    ADMXRC_CleanupDMA(devCtx);
    ADMXRC_CleanupDMATargets(devCtx);
    devCtx->owner.handle = NULL;
  }

  return status;
}

static OS_STATUS
onMap(
  OS_DEVICE_OBJECT*   devObj,
  OS_MAPPABLE_REGION* region,
  void*               context,
  void*               virtual,
  os_uint64           phys,
  os_uint64           length)
{
  ADMXRC* devCtx;
  int     index = (int) context;

  os_dbgprint(2, ("onMap: entered, context = 0x%p virtual = 0x%p phys=0x%8.8lx%8.8lx length=0x%8.8lx%8.8lx\n",
		  context,
		  virtual,
		  (unsigned long) os_uint64_hi(phys), (unsigned long) os_uint64_lo(phys),
		  (unsigned long) os_uint64_hi(length), (unsigned long) os_uint64_lo(length)));

  devCtx = os_get_device_context(devObj);

  devCtx->info.spaces[index].VirtualBase = virtual;
  devCtx->info.spaces[index].VirtualSize = (unsigned long) length;

  return OS_STATUS_SUCCESS;
}

static OS_STATUS
onIoctl(
  OS_DEVICE_OBJECT* devObj,
  OS_USER_CLIENT    client,
  OS_IO_REQUEST*    req)
{
  ADMXRC*           devCtx;
  OS_STATUS         status = OS_STATUS_SUCCESS;
  OS_SPINLOCK_FLAGS f;
  os_boolean        handled;
  os_boolean        owner;

  os_dbgprint(4, ("onIoctl: entered, client = 0x%p code = 0x%lx(%ld)\n",
		  client,
		  (unsigned long) req->param.code, (unsigned long) req->param.code));

  devCtx = os_get_device_context(devObj);

  os_acquire_spinlock(&devCtx->owner.lock, &f);
  owner = (client == devCtx->owner.handle);
  os_release_spinlock(&devCtx->owner.lock, f);

  handled = TRUE; /* Assume it will be handled in the next switch statement */
  switch (req->param.code) {
  case ADMXRC_IOCTL_OPENCARD:
    status = ADMXRC_OpenCard(devCtx, client, req);
    break;
  case ADMXRC2_IOCTL_OPENCARD:
    status = ADMXRC2_OpenCard(devCtx, client, req);
    break;
  case ADMXRC2_IOCTL_OPENCARDBYINDEX:
    status = ADMXRC2_OpenCardByIndex(devCtx, client, req);
    break;
  case ADMXRC2_IOCTL_GETBANKINFO:
    status = ADMXRC2_GetBankInfo(devCtx, req);
    break;
  case ADMXRC_IOCTL_GETCARDINFO:
    status = ADMXRC_GetCardInfo(devCtx, req, owner);
    break;
  case ADMXRC2_IOCTL_GETCARDINFO:
    status = ADMXRC2_GetCardInfo(devCtx, req);
    break;
  case ADMXRC_IOCTL_GETCLOCKTYPE:
    status = ADMXRC_GetClockType(devCtx, req);
    break;
  case ADMXRC2_IOCTL_GETSPACEINFO:
    status = ADMXRC2_GetSpaceInfo(devCtx, req, owner);
    break;
  case ADMXRC_IOCTL_GETVERSIONINFO:
    status = ADMXRC_GetVersionInfo(devCtx, req);
    break;
  default:
    /*
    ** The IOCTL wasnt handled. Try the switch statement below,
    ** if the calling thread owns the card.
    */
    handled = FALSE;
    break;
  }
  
  if (handled) {
    goto finish;
  }

  if (!owner) {
    req->status.status = OS_STATUS_CANNOTSHARE;
    os_complete_request(req);
    return status;
  }

  switch (req->param.code) {
  case ADMXRC2_IOCTL_CANCELWAITFORINTERRUPT:
    status = ADMXRC_CancelWaitForInterrupt(devCtx, req);
    break;
  case ADMXRC_IOCTL_DODMA:
    status = ADMXRC_DoDMA(devCtx, req);
    break;
  case ADMXRC_IOCTL_DODMAIMMEDIATE:
    status = ADMXRC_DoDMAImmediate(devCtx, req);
    break;
  case ADMXRC_IOCTL_LOADFPGA:
    status = ADMXRC_LoadFPGA(devCtx, req);
    break;
  case ADMXRC_IOCTL_LOADFPGADMA:
    status = ADMXRC_LoadFPGADMA(devCtx, req);
    break;
  case ADMXRC_IOCTL_MAPDIRECTMASTER:
    status = ADMXRC_MapDirectMaster(devCtx, req);
    break;
  case ADMXRC_IOCTL_PROGRAMBYTES:
    status = ADMXRC_ProgramBytes(devCtx, req);
    break;
  case ADMXRC_IOCTL_READ:
    status = ADMXRC_Read(devCtx, req);
    break;
  case ADMXRC2_IOCTL_READCONFIG:
    status = ADMXRC2_ReadConfig(devCtx, req);
    break;
  case ADMXRC_IOCTL_READREG:
    status = ADMXRC_ReadReg(devCtx, req);
    break;
  case ADMXRC_IOCTL_SETCLOCK:
    status = ADMXRC_SetClock(devCtx, req);
    break;
  case ADMXRC2_IOCTL_SETCLOCK:
    status = ADMXRC2_SetClock(devCtx, req);
    break;
  case ADMXRC_IOCTL_SETUPDMA:
    status = ADMXRC_SetupDMA(devCtx, req);
    break;
  case ADMXRC_IOCTL_SYNCDIRECTMASTER:
    status = ADMXRC_SyncDirectMaster(devCtx, req);
    break;
  case ADMXRC_IOCTL_UNSETUPDMA:
    status = ADMXRC_UnsetupDMA(devCtx, req);
    break;
  case ADMXRC_IOCTL_WRITE:
    status = ADMXRC_Write(devCtx, req);
    break;
  case ADMXRC_IOCTL_WAITFORINTERRUPT:
    status = ADMXRC_WaitForInterrupt(devCtx, req);
    break;
  case ADMXRC2_IOCTL_WRITECONFIG:
    status = ADMXRC2_WriteConfig(devCtx, req);
    break;
  case ADMXRC_IOCTL_WRITEFLASH:
    status = ADMXRC_WriteFlash(devCtx, req);
    break;
  case ADMXRC_IOCTL_WRITEREG:
    status = ADMXRC_WriteReg(devCtx, req);
    break;
  default:
    req->status.status = OS_STATUS_INVALIDIOCTL;
    os_complete_request(req);
    return status;
  }

 finish:
  if (!OS_PENDING(status)) {
    req->status.status = status;
    os_complete_request(req);
  }

  return status;
}

/*
** -----------------------------------------------------------------------------
** Required by framework
** -----------------------------------------------------------------------------
*/

const char* driver_name = DRIVER_NAME;

OS_STATUS
driver_init(
  OS_DRIVER_OBJECT* drvObj)
{
  os_dbgprint(1, ("driver_init: entered\n"));

  devDescs[0].pci.vId = PCI_PLX9080_VENDOR_ID;
  devDescs[0].pci.dId = PCI_PLX9080_DEVICE_ID;
  devDescs[0].pci.subVId = PCI_SUBVENDOR_ID_ADMXRC;
  devDescs[0].pci.subDId = PCI_SUBDEVICE_ID_ADMXRC;
  devDescs[0].type = OS_DEVICE_DESC_TYPE_PCI;
  devDescs[1].pci.vId = PCI_PLX9080_VENDOR_ID;
  devDescs[1].pci.dId = PCI_PLX9080_DEVICE_ID;
  devDescs[1].pci.subVId = PCI_SUBVENDOR_ID_ADMXRC2_L;
  devDescs[1].pci.subDId = PCI_SUBDEVICE_ID_ADMXRC2_L;
  devDescs[1].type = OS_DEVICE_DESC_TYPE_PCI;
  devDescs[2].pci.vId = PCI_PLX9656_VENDOR_ID;
  devDescs[2].pci.dId = PCI_PLX9656_DEVICE_ID;
  devDescs[2].pci.subVId = PCI_SUBVENDOR_ID_ADMXRC2;
  devDescs[2].pci.subDId = PCI_SUBDEVICE_ID_ADMXRC2;
  devDescs[2].type = OS_DEVICE_DESC_TYPE_PCI;
  devDescs[3].type = OS_DEVICE_DESC_TYPE_NONE;
  drvObj->hardwareIds = devDescs;
  drvObj->drvName = DRIVER_NAME;
  drvObj->onCreateDevice = onCreateDevice;
  drvObj->onStartDevice = onStartDevice;
  drvObj->onStopDevice = onStopDevice;
  drvObj->onDeleteDevice = onDeleteDevice;
  drvObj->onOpen = onOpen;
  drvObj->onClose = onClose;
  drvObj->onMap = onMap;
  drvObj->onIoctl = onIoctl;

  return 0;
}

void
driver_uninit(
  OS_DRIVER_OBJECT* drvObj)
{
  os_dbgprint(1, ("driver_uninit: entered\n"));
}
