/*
** platform.c
**
** Platform abstraction layer
*/

#include "platform.h"

#if !defined(OS_ALLOC_KIOVEC_SZ)
#   define alloc_kiovec_sz(x, y, z) alloc_kiovec(x, y)
#   define free_kiovec_sz(x, y, z) free_kiovec(x, y)
#endif

#if (LINUX_VERSION_CODE >= 0x020400)
#   define _os_page_is_high(k, i) PageHighMem((k)->maplist[i]);
#else
#   define _os_page_is_high(k, i) PageHighMem((k)->pagelist[i]);
#endif

int _os_dbglevel = 0;          /* no debugging messages by default          */

static void
_os_isr(
  int             irq,
  void*           context,
  struct pt_regs* regs)
{
  OS_ATTACHED_IRQ* irqObj = (OS_ATTACHED_IRQ*) context;

  irqObj->isr(irqObj->deviceContext);
}

static void
_os_dpc_handler(
  unsigned long context)
{
  OS_DEVICE_OBJECT* devObj;
  OS_DPC*           dpc = (OS_DPC*) context;
  OS_SPINLOCK_FLAGS f;
  void*             arg;
  void*             dpc_context;
  OS_DPC_FUNC*      handler;
  os_boolean        destroy;

  devObj = dpc->devObj;
  os_acquire_spinlock(&devObj->dpcLock, &f);
  dpc->scheduled = FALSE;
  handler = dpc->handler;
  dpc_context = dpc->context;
  arg = dpc->argument;
  destroy = dpc->destroy;
  os_release_spinlock(&devObj->dpcLock, f);

  handler(dpc_context, arg);

  if (destroy) {
    wake_up(dpc->wake);
  }
}

/*
** -------------------------------------------------------------------
** Exported to client code
** -------------------------------------------------------------------
*/

OS_STATUS
os_allow_mapping(
  OS_DEVICE_OBJECT*    devObj,
  os_uint64            phys,
  os_uint64            length,
  void*                context,
  OS_MAPPABLE_REGION** handle)
{
  OS_MAPPABLE_REGION* region;

  os_dbgprint(3, ("os_allow_mapping: phys = 0x%8.8lx%8.8lx length = 0x%8.8lx%8.8lx context = 0x%p\n",
		  (unsigned long) os_uint64_hi(phys),
		  (unsigned long) os_uint64_lo(phys),
		  (unsigned long) os_uint64_hi(length),
		  (unsigned long) os_uint64_lo(length),
		  context));

  region = (OS_MAPPABLE_REGION*) os_kernel_malloc(sizeof(OS_MAPPABLE_REGION));
  if (region == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  region->devObj = devObj;
  region->phys = phys;
  region->length = length;
  region->context = context;
  region->next = devObj->maps.regions;
  region->prev = &devObj->maps.regions;
  devObj->maps.regions = region;

  os_dbgprint(3, ("os_allow_mapping: created region 0x%p\n", region));

  *handle = region;

  return OS_STATUS_SUCCESS;
}

OS_STATUS
os_attach_interrupt(
  OS_DEVICE_OBJECT* devObj,
  OS_ISR*           isr,
  long              vector,
  const char*       devName,
  OS_ATTACHED_IRQ** o)
{
  OS_ATTACHED_IRQ* obj;
  int              err;

  obj = os_kernel_malloc(sizeof(OS_ATTACHED_IRQ));
  if (obj == NULL) {
    return OS_STATUS_NOMEMORY;
  }
  obj->vector = vector;
  obj->isr = isr;
  obj->devObj = devObj;
  obj->deviceContext = devObj->deviceContext;

  /* SA_INTERRUPT => fast handler, so interrupts are disabled in ISR */
  err = request_irq(obj->vector, _os_isr, SA_SHIRQ | SA_INTERRUPT, devName, (void*) obj);
  if (err == 0) {
    *o = obj;
    return OS_STATUS_SUCCESS;
  } else {
    os_kernel_free(obj);
    if (err == -EBUSY) {
      return OS_STATUS_CANNOTSHARE;
    } else {
      return OS_STATUS_UNSUCCESSFUL;
    }
  }
}

void
os_complete_request(
  OS_IO_REQUEST* req)
{
  OS_DEVICE_OBJECT* devObj;
  OS_SPINLOCK_FLAGS f;

  os_dbgprint(4, ("os_complete_request: req = 0x%p, status = %lx, length = %ld\n",
		  (void*) req, (unsigned long) req->status.status, (unsigned long) req->status.length));

  devObj = req->opaque.devObj;
  os_acquire_spinlock(&devObj->cancelLock, &f);
  req->opaque.completed = TRUE;
  wake_up(req->opaque.wait_q);
  os_release_spinlock(&devObj->cancelLock, f);
}

OS_STATUS
os_create_dpc(
  OS_DEVICE_OBJECT* devObj,
  OS_DPC_FUNC*      handler,
  void*             context,
  OS_DPC**          dpc)
{
  OS_DPC* p;

  p = (OS_DPC*) os_kernel_malloc(sizeof(OS_DPC));
  if (p == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  p->devObj = devObj;
  p->handler = handler;
  p->context = context;
  p->destroy = FALSE;
  p->scheduled = FALSE;
  *dpc = p;

  return OS_STATUS_SUCCESS;
}

OS_STATUS
os_create_lcache(
  unsigned long item_size,
  int           flags,
  OS_LCACHE**   cache)
{
  static int    n = 0;
  char          name[20];
  unsigned long f;
  OS_LCACHE*    c;

  sprintf(name, "os_unnamed_lcache%d", n);
  n++;
  f = SLAB_HWCACHE_ALIGN | SLAB_CACHE_DMA;
  c = kmem_cache_create(name, item_size, 0, f, NULL, NULL);
  if (c == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  *cache = c;
  return OS_STATUS_SUCCESS;
}

OS_STATUS
os_create_usbd(
  void*         user,
  unsigned long length,
  OS_USBD**     usbd)
{
  OS_STATUS      status;
  OS_USBD*       u = NULL;
  int            err;
  os_boolean     mapped = FALSE;

  u = (OS_USBD*) os_kernel_malloc(sizeof(OS_USBD));
  if (u == NULL) {
    status = OS_STATUS_NOMEMORY;
    goto failed;
  }

  u->kio[0] = NULL;
  u->nbhs[0] = KIO_MAX_SECTORS;
  err = alloc_kiovec_sz(1, u->kio, u->nbhs);
  if (err) {
    status = OS_STATUS_NOMEMORY;
    goto failed;
  }

  err = map_user_kiobuf(READ, u->kio[0], (unsigned long) user, length);
  if (err) {
    status = OS_STATUS_UNSUCCESSFUL;
    goto failed;
  }
  mapped = TRUE;

#if 0
  /*
  ** EXPERIMENTAL
  **
  ** Here, check that every page in the user buffer is in
  ** DMA accessible memory, and fail if not.
  */
  {
    int            i, n;

    n = u->kio[0]->nr_pages;
    for (i = 0; i < n; i++) {
      if (_os_page_is_high(u->kio[0], i)) {
	/* Page is in high memory - we can't use it */
	status = OS_STATUS_UNSUCCESSFUL;
	goto failed;
      }
    }
  }
#endif

  *usbd = u;

  return OS_STATUS_SUCCESS;

 failed:
  if (u != NULL) {
    if (mapped) {
      unmap_kiobuf(u->kio[0]);
    }
    if (u->kio != NULL) {
      free_kiovec_sz(1, u->kio, u->nbhs);
    }
    os_kernel_free(u);
  }
  return status;
}

os_boolean
os_delay_thread(
  unsigned long* ticks)
{
  long          start = jiffies;
  long          elapsed;
  unsigned long n = *ticks;

  set_current_state(TASK_INTERRUPTIBLE);
  schedule_timeout(n);

  /* Calculate remaining time */
  elapsed = (unsigned long) (jiffies - start);
  if (elapsed > n) {
    *ticks = 0;
  } else {
    *ticks = n - elapsed;
  }

  /* return TRUE if interrupted by a signal, otherwise FALSE */
  return (os_boolean) signal_pending(current);
}

void
os_destroy_dpc(
  OS_DPC* dpc)
{
  OS_DEVICE_OBJECT* devObj;
  OS_SPINLOCK_FLAGS f;
  os_boolean        wait_destroy;
  wait_queue_head_t wait_q;
  wait_queue_t      wait;

  if (dpc == NULL) {
    os_dbgprint(0, ("os_destroy_dpc: NULL pointer\n"));
    return;
  }

  devObj = dpc->devObj;
  os_acquire_spinlock(&devObj->dpcLock, &f);
  if (dpc->scheduled) {
    /*
    ** DPC is either scheduled or running - we must wait for
    ** it to finish before freeing the control structure
    */

    wait_destroy = TRUE;
    set_current_state(TASK_UNINTERRUPTIBLE);
    init_waitqueue_head(&wait_q);
    init_waitqueue_entry(&wait, current);
    add_wait_queue(&wait_q, &wait);
    dpc->wake = &wait_q;
    dpc->destroy = TRUE;
  } else {
    wait_destroy = FALSE;
  }
  os_release_spinlock(&devObj->dpcLock, f);

  if (wait_destroy) {
    schedule();
    remove_wait_queue(&wait_q, &wait);
  }

  os_kernel_free(dpc);
}

void
os_destroy_lcache(
  OS_LCACHE* cache)
{
  if (cache == NULL) {
    os_dbgprint(0, ("os_destroy_lcache: NULL pointer\n"));
    return;
  }

  if (kmem_cache_destroy(cache)) {
    os_dbgprint(0, ("os_destroy_lcache: memory leak detected\n"));
  }
}

void
os_destroy_usbd(
  OS_USBD* usbd)
{
  unmap_kiobuf(usbd->kio[0]);
  free_kiovec_sz(1, usbd->kio, usbd->nbhs);
  os_kernel_free(usbd);
}

void
os_detach_interrupt(
  OS_ATTACHED_IRQ* obj)
{
  free_irq(obj->vector, obj);
  os_kernel_free(obj);
}

void
os_disallow_mapping(
  OS_DEVICE_OBJECT*   devObj,
  OS_MAPPABLE_REGION* region)
{
  os_dbgprint(3, ("os_disallow_mapping: region = 0x%p\n", region));

  *region->prev = region->next;

  os_kernel_free(region);
}

void
os_flush_io_buffer(
  OS_USBD*      usbd,
  unsigned long offset,
  unsigned long length,
  os_boolean    write)
{
  if (write) {
    /* What goes here? */
  } else {
    /* What goes here? */
  }
}

void
os_init_event(
  OS_EVENT* event)
{
  os_init_spinlock(&event->lock);
  init_waitqueue_head(&event->q);
  event->raised = FALSE;
}

void*
os_invoke_atomic(
  OS_ATTACHED_IRQ* obj,
  OS_INVOKE_FUNC*  function,
  void*            context)
{
  void*             ret;

  disable_irq(obj->vector);
  ret = function(context);
  enable_irq(obj->vector);

  return ret;
}

os_uint64
os_kernel_to_bus(
  OS_DEVICE_OBJECT* devObj,
  void*             kern)
{
  return (os_uint64) ((unsigned long) virt_to_bus(kern));
}

void*
os_map_kernel(
  os_uint64 phys,
  os_uint64 length)
{
  void* ptr;

  if (!length) {
    os_dbgprint(0, ("os_map_kernel: attempt to map zero-length region\n"));
  }

  ptr = ioremap_nocache(phys, length);
  
  return ptr;
}

void
os_unmap_kernel(
  void*     kern,
  os_uint64 phys,
  os_uint64 length)
{
  if (kern == NULL) {
    os_dbgprint(0, ("os_unmap_kernel: NULL pointer\n"));
  }

  iounmap(kern);
}

OS_STATUS
os_schedule_dpc(
  OS_DPC* dpc,
  void*   argument)
{
  OS_DEVICE_OBJECT* devObj;
  OS_SPINLOCK_FLAGS f;

  devObj = dpc->devObj;
  os_acquire_spinlock(&devObj->dpcLock, &f);
  if (dpc->scheduled) {
    /* Already scheduled, don't schedule again */
    os_release_spinlock(&devObj->dpcLock, f);
    return OS_STATUS_SUCCESS;
  }
  if (dpc->destroy) {
    /* About to be destroyed - this indicates a programming error */
    os_release_spinlock(&devObj->dpcLock, f);
    os_dbgprint(0, ("os_schedule_dpc: DPC scheduled when being destroyed\n"));
    return OS_STATUS_UNSUCCESSFUL;
  }
  dpc->scheduled = TRUE;
  os_release_spinlock(&devObj->dpcLock, f);

  tasklet_init(&dpc->task_queue, _os_dpc_handler, (unsigned long) dpc);
  dpc->argument = argument;
  tasklet_schedule(&dpc->task_queue);
  mark_bh(IMMEDIATE_BH);

  return OS_STATUS_SUCCESS;
}

os_boolean
os_set_cancel_routine(
  OS_IO_REQUEST*    req,
  OS_IO_CANCELFUNC* func)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&req->opaque.devObj->cancelLock, &f);
  if (req->opaque.cancel) {
    /*
    ** Cannot change the cancel routine, as if set, it may be already
    ** running or about to run
    */
    os_release_spinlock(&req->opaque.devObj->cancelLock, f);
    return FALSE;
  } else {
    /*
    ** Can successfully change the cancel routine as if set, it's not
    ** running yet.
    */
    req->opaque.cancelFunc = func;
    os_release_spinlock(&req->opaque.devObj->cancelLock, f);
    return TRUE;
  }
}

void
os_signal_event(
  OS_EVENT* event)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&event->lock, &f);
  event->raised = TRUE;
  wake_up(&event->q);
  os_release_spinlock(&event->lock, f);
}

os_boolean
os_wait_event(
  OS_EVENT*  event,
  os_boolean interruptible)
{
  OS_SPINLOCK_FLAGS f;
  wait_queue_t      wait;

  os_acquire_spinlock(&event->lock, &f);
  if (event->raised) {
    os_release_spinlock(&event->lock, f);
    return FALSE;
  }

  init_waitqueue_entry(&wait, current);
  set_current_state(interruptible ? TASK_INTERRUPTIBLE : TASK_UNINTERRUPTIBLE);
  add_wait_queue(&event->q, &wait);
  os_release_spinlock(&event->lock, f);
  schedule();
  if (interruptible) {
    if (signal_pending(current)) {
      /* Caught a signal */
      return TRUE;
    }
  }

  /* No signal caught */
  return FALSE;
}
