/*
** rw.c     
**
** IOCTL handlers for reading and writing the FPGA, control registers and Flash.
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "rw.h"

/*
** Validate the parameters of an IoRead or IoWrite call and
** return a pointer to the kernel address where the transfer
** is to start.
*/
static OS_STATUS
validate_rw(
  ADMXRC*       devCtx,
  unsigned long flags,
  unsigned long width,
  unsigned long length,
  os_uint32     local,
  void**        devicemem)
{
  unsigned long span;
  os_uint32     maxlocal;
  os_boolean    fixed = (flags & ADMXRC2_IOFIXED) != 0;
  os_uint8*     mem;
	
  span = fixed ? width : length;
	
  if (flags & ADMXRC2_IOADAPTER) {
    if (local >= devCtx->info.regsLimit) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
    mem = (os_uint8*) devCtx->hw.regs.untyped;
    maxlocal = devCtx->info.regsLimit;
  } else {
    if ((local >= devCtx->info.spaces[0].LocalBase) &&
	(local < (devCtx->info.spaces[0].LocalBase + devCtx->info.spaces[0].LocalSize))) {
      /* PLX S0 */
      local -= devCtx->info.spaces[0].LocalBase;
      mem = (os_uint8*) devCtx->hw.fpga;
      maxlocal = devCtx->info.spaces[0].LocalSize;
    } else if ((local >= devCtx->info.spaces[1].LocalBase) &&
	       (local < (devCtx->info.spaces[1].LocalBase + devCtx->info.spaces[1].LocalSize))) {
      /* PLX S1 */
      local -= devCtx->info.spaces[1].LocalBase;
      mem = (os_uint8*) devCtx->hw.control;
      maxlocal = devCtx->info.spaces[1].LocalSize;
    } else {
      /* Invalid local bus address */
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
  }
	
  /* Don't permit the space boundary to be crossed */
  if (local + span > maxlocal) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
	
  /* Get the system address at which to begin writing */
  mem += local;
  
  *devicemem = mem;
  
  return OS_STATUS_SUCCESS;
}

/*
**	Write data to the ADMXRC Flash.
**
**	Input buffer consists of data to be written to the EPROM. Assumes device is erased.
**	Structure of buffer is :-
**	
**		ULONG Offset in bytes from base address
**		ULONG Length in bytes of data to be written
**		ULONG Flags
**		UCHAR CPLD load flag (bit 0=0 to load-on-reset)
**		UCHAR data[]
**
**	The total length is 4+4+4+length bytes.
*/
OS_STATUS
ADMXRC_WriteFlash(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
}

/*
** ADMXRC_ProgramBytes() interprets a simple program and is intended
** to enable flash programming algorithms to be implemented at
** application level with a reduced number of IOCTLs.
**
** The program format is an array of pairs of ULONGs:
**
**       +----+--------------------------------------------+
**       | op |                 flash_addr                 |
**       +----+--------------------------------------------+
**       |                    operands                     |
**       +-------------------------------------------------+
**
** where
**       flash_addr = bits [27:0] of word 0
**       op = bits [31:28] of word 0
**
** There are currently 3 ops defined:
**
**       op = 0  => write to flash
**       op = 1  => wait for flash location to read back value
**       op = 2  => read flash
**
** Since flash_addr is 28 bits wide, Flash up to 256MB in size is supported.
**
** OP 0 (write value)
** ------------------
**
**       +----+--------------------------------------------+
**       |  0 |                 flash_addr                 |
**       +----+------------------+------------+------------+
**       |                       |    mask    |    data    |
**       +-----------------------+------------+------------+
**
** where
**       data = bits[7:0] of word 1
**       mask = bits[15:8] of word 1
**
** the operation performed is:
**
**    FLASH_WRITE(flash_addr, (FLASH_READ(flash_addr) & mask) | data);
**
**
** OP 1 (wait for value)
** ---------------------
**
**       +----+--------------------------------------------+
**       |  1 |                 flash_addr                 |
**       +-+--+------------------+------------+------------+
**       |m|      timeout        |    mask    |    data    |
**       +-+---------------------+------------+------------+
**
** where
**       data = bits[7:0] of word 1
**       mask = bits[15:8] of word 1
**		 timeout = bits[30:16] of word 1
**		 m = bits[31] of word 1
**
** the operation performed is:
**
**    while ((FLASH_READ(flash_addr) & mask) != data) { };
**
** There is a timeout on this wait; the timeout is measured in
** milliseconds if m == 1 and in microseconds if m == 0. If the
** timeout period elapses without the specified value being read,
** the return value from AdmXrcProgramBytes() is ADMXRC_TIMEOUT.
**
** OP 2 (read value)
** -----------------
**
**       +----+--------------------------------------------+
**       |  2 |                 flash_addr                 |
**       +----+-------------------------------+------------+
**       |                                    |    data    |
**       +------------------------------------+------------+
**
** where
**       data = bits[7:0]
**
** the operation performed is:
**
**    data = FLASH_READ(flash_addr);
**
*/
OS_STATUS
ADMXRC_ProgramBytes(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  unsigned long               i, length;
  os_uint32*                  program;
  os_uint32                   op1, op2, addr, data, mask;
  os_uint8                    val;
  os_uint8                    page;

  os_dbgprint(3, ("ADMXRC_ProgramBytes: entered\n"));

  length = req->param.massSize;
  program = req->param.massBuf;
  if ((length & (2 * sizeof(os_uint32) - 1)) ||
      (length && program == NULL)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Interpret the programming codes supplied */
  for (i = 0; i < length; i += 2 * sizeof(os_uint32)) {
    op1 = program[0];
    op2 = program[1];
    addr = op1 & 0x001fffffU;
    page = (os_uint8) ((op1 >> 21) & 0x7fU);
    data = (os_uint8) (op2 >> 0);
    mask = (os_uint8) (op2 >> 8);
    
    if (devCtx->info.card.BoardType == ADMXRC2_BOARD_ADMXRC2) {
      /* Set Flash page register on ADM-XRCII only */
      os_dbgprint(3, ("ADMXRC_ProgramBytes: setting page register to 0x%2.2lx\n",
		      (unsigned long) page));
      os_write_register_8(&devCtx->hw.control->u.FLPAGE, page);
    }

    switch ((op1 >> 28) & 0xfU) {
    case 0x0:
      {
	os_dbgprint(3, ("ADMXRC_ProgramBytes: op 0 mask = 0x%2.2lx data = 0x%2.2lx\n",
			(unsigned long) mask, (unsigned long) data));

	if (mask) {
	  /* Write a value to flash location with mask */
	  val = os_read_register_8(devCtx->hw.flash + addr);
	  data |= val & mask;
	}
	os_write_register_8(devCtx->hw.flash + addr, data);
	break;
      }
    case 0x1:
      {
	/* Wait for a certain flash location to read a certain value */
	os_boolean    done = FALSE;
	unsigned long timeout;
	unsigned long expires, dummy;

	os_dbgprint(3, ("ADMXRC_ProgramBytes: op 1 mask = 0x%2.2lx data = 0x%2.2lx\n",
			(unsigned long) mask, (unsigned long) data));
	
	/* Get timeout value */
	timeout = (op2 >> 16) & 0x7fffU;

	if (op2 & (1 << 31)) {
	  /* Timeout is in milliseconds */
	  expires = timeout * os_ticks_per_second / 1000;
	  if (!expires) {
	    expires++;
	  }
	  expires += os_get_time();
	  
	  done = FALSE;
	  while (1) {
	    val = os_read_register_8(devCtx->hw.flash + addr);
	    if ((val & mask) == data) {
	      done = TRUE;
	      break;
	    }

	    dummy = 1;
	    if (os_delay_thread(&dummy)) {
	      return OS_MAKE_STATUS(ADMXRC_CANCELLED);
	    }
	    
	    if (os_get_time() >= expires) {
	      break;
	    }
	  }
	} else {
	  /* Timeout is in microseconds */
	  while (timeout) {
	    unsigned long chunk = (timeout > 5) ? 5 : timeout;
	    
	    val = os_read_register_8(devCtx->hw.flash + addr);
	    if ((val & mask) == data) {
	      done = TRUE;
	      break;
	    }

	    os_delay_busy_us(chunk);
	    timeout -= chunk;
	  }
	}

	if (!done) {
	  return OS_MAKE_STATUS(ADMXRC_TIMEOUT);
	}
      }
      break;
    case 0x2:
      {
	/* Read value from flash location */
	val = os_read_register_8(devCtx->hw.flash + addr);
	program[1] &= ~0xffU;
	program[1] |= val;

	os_dbgprint(3, ("ADMXRC_ProgramBytes: op 2 addr = 0x%lx data = 0x%2.2lx\n",
			(unsigned long) addr, (unsigned long) val));
      }
      break;
    }
    program += 2;
  }

  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_WriteReg(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_WRITEREG* iptr;
  
  iptr = (ADMXRC_IOCTLS_WRITEREG*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  
  os_write_register_8(((os_uint8*) devCtx->hw.control) + iptr->in.index, iptr->in.value);
  
  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_ReadReg(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_READREG* iptr;
  ADMXRC_IOCTLS_READREG* optr;
  
  iptr = (ADMXRC_IOCTLS_READREG*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_READREG*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  optr->out.value = os_read_register_8(((os_uint8*) devCtx->hw.control) + iptr->in.index);
  req->status.length = sizeof(optr->out);

  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_Write(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_WRITE* iptr;
  OS_STATUS            status;
  void*                devicemem;
  os_boolean           fixed, endian;
  unsigned long        length;
  void*                data;

  iptr = (ADMXRC_IOCTLS_WRITE*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  data = (void*) req->param.massBuf;
  length = req->param.massSize;

  /* Only byte (0), word (1) or longword (2) accesses are allowed */
  if (iptr->in.width > ADMXRC_IOLONG) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Check the local bus address, transfer size and flags parameters */
  status = validate_rw(devCtx,
		       iptr->in.flags,
		       1 << iptr->in.width,
		       length,
		       iptr->in.local,
		       &devicemem);
  if (status != OS_STATUS_SUCCESS) {
    return status;
  }

  /*
  ** Note that 'length' is always in bytes.
  ** Transfers of SHORT and LONG will be truncated if not correct multiple
  */
  fixed = (iptr->in.flags & ADMXRC_IOFIXED) != 0;
  endian = (iptr->in.flags & ADMXRC2_IOADAPTER) != 0;

  if (endian) {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_write_buffer_8endian(devicemem, data, length);
	break;
      case ADMXRC_IOWORD:
	os_write_buffer_16endian(devicemem, data, length);
	break;
      case ADMXRC_IOLONG:
	os_write_buffer_32endian(devicemem, data, length);
	break;
      }
    } else {
      
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8*  puc;
	  
	  puc = (os_uint8*) data;
	  while (length > 0) {
	    os_write_register_8endian((os_uint8*) devicemem, *puc);
	    puc++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) data;
	  length >>= 1;
	  while (length > 0) {
	    os_write_register_16endian((os_uint16*) devicemem, *pus);
	    pus++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) data;
	  length >>= 2;
	  while (length > 0) {
	    os_write_register_32endian((os_uint32*) devicemem, *pul);
	    pul++;
	    length--;
	  }
	  break;
	}
      }
    }
  } else {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_write_buffer_8(devicemem, data, length);
	break;
      case ADMXRC_IOWORD:
	os_write_buffer_16(devicemem, data, length);
	break;
      case ADMXRC_IOLONG:
	os_write_buffer_32(devicemem, data, length);
	break;
      }
    } else {
      
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8*  puc;
	  
	  puc = (os_uint8*) data;
	  while (length > 0) {
	    os_write_register_8((os_uint8*) devicemem, *puc);
	    puc++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) data;
	  length >>= 1;
	  while (length > 0) {
	    os_write_register_16((os_uint16*) devicemem, *pus);
	    pus++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) data;
	  length >>= 2;
	  while (length > 0) {
	    os_write_register_32((os_uint32*) devicemem, *pul);
	    pul++;
	    length--;
	  }
	  break;
	}
      }
    }
  }
  
  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_Read(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_READ* iptr;
  OS_STATUS           status;
  void*               devicemem;
  os_boolean          fixed, endian;
  unsigned long       length;
  void*               buffer;

  iptr = (ADMXRC_IOCTLS_READ*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  buffer = (void*) req->param.massBuf;
  length = req->param.massSize;

  /* Only byte (0), word (1) or longword (2) accesses are allowed */
  if (iptr->in.width > ADMXRC_IOLONG) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* Check the local bus address, transfer size and flags parameters */
  status = validate_rw(devCtx,
		       iptr->in.flags,
		       1 << iptr->in.width,
		       length,
		       iptr->in.local,
		       &devicemem);
  if (status != OS_STATUS_SUCCESS) {
    return status;
  }

  /*
  ** Note that 'length' is always in bytes.
  ** Transfers of 16 bits and 32 bits will be truncated if not correct multiple
  */

  fixed = (iptr->in.flags & ADMXRC2_IOFIXED) != 0;
  endian = (iptr->in.flags & ADMXRC2_IOADAPTER) != 0;

  if (endian) {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_read_buffer_8endian(buffer, (void*) devicemem, length);
	break;
      case ADMXRC_IOWORD:
	os_read_buffer_16endian(buffer, (void*) devicemem, length);
	break;
      case ADMXRC_IOLONG:
	os_read_buffer_32endian(buffer, (void*) devicemem, length);
	break;
      }
    } else {		
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8* puc;
	  
	  puc = (os_uint8*) buffer;
	  while (length > 0) {
	    *puc = os_read_register_8endian((os_uint8*) devicemem);
	    puc++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) buffer;
	  length >>= 1;
	  while (length > 0) {
	    *pus = os_read_register_16endian((os_uint16*) devicemem);
	    pus++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) buffer;
	  length >>= 2;
	  while (length > 0) {
	    *pul = os_read_register_32endian((os_uint32*) devicemem);
	    pul++;
	    length--;
	  }
	  break;
	}
      }
    }
  } else {
    if (!fixed) {
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	os_read_buffer_8(buffer, (void*) devicemem, length);
	break;
      case ADMXRC_IOWORD:
	os_read_buffer_16(buffer, (void*) devicemem, length);
	break;
      case ADMXRC_IOLONG:
	os_read_buffer_32(buffer, (void*) devicemem, length);
	break;
      }
    } else {		
      switch (iptr->in.width) {
      case ADMXRC_IOBYTE:
	{
	  os_uint8* puc;
	  
	  puc = (os_uint8*) buffer;
	  while (length > 0) {
	    *puc = os_read_register_8((os_uint8*) devicemem);
	    puc++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOWORD:
	{
	  os_uint16* pus;
	  
	  pus = (os_uint16*) buffer;
	  length >>= 1;
	  while (length > 0) {
	    *pus = os_read_register_16((os_uint16*) devicemem);
	    pus++;
	    length--;
	  }
	  break;
	}
      case ADMXRC_IOLONG:
	{
	  os_uint32* pul;
	  
	  pul = (os_uint32*) buffer;
	  length >>= 2;
	  while (length > 0) {
	    *pul = os_read_register_32((os_uint32*) devicemem);
	    pul++;
	    length--;
	  }
	  break;
	}
      }
    }
  }
  
  return OS_STATUS_SUCCESS;
}
