/*
** setclock.c     
**
** IOCTL handlers for setting the clock generators
*/

#include "platform.h"
#include "device.h"
#include "ioctl.h"
#include "setclock.h"

static const unsigned long mclkRange[16][2] = {
  { 49990000,  51000000 },
  { 50990000,  53200000 },
  { 53190000,  58500000 },
  { 58490000,  60700000 },
  { 60690000,  64400000 },
  { 60390000,  66800000 },
  { 66790000,  73500000 },
  { 73490000,  75600000 },
  { 75590000,  80900000 },
  { 80890000,  83200000 },
  { 83190000,  91500000 },
  { 91490000, 100000000 },
  { 99990000, 120100000 },
  { 99990000, 120100000 },
  { 99990000, 120100000 },
  { 99990000, 120100000 }
};

static void*
programICD2061Atomic(
  void* context)
{
  ADMXRC*          devCtx = (ADMXRC*) context;
  os_uint32        clockword;
  int              i;
  os_uint8         d;

  clockword = devCtx->setclock.clockword;

  os_write_register_8(&devCtx->hw.control->CCON, 3);
  
  for (i = 0; i < 6; i++) {
    os_write_register_8(&devCtx->hw.control->CCON, 2);
    os_write_register_8(&devCtx->hw.control->CCON, 3);
  }
  
  os_write_register_8(&devCtx->hw.control->CCON, 2);
  os_write_register_8(&devCtx->hw.control->CCON, 0);
  os_write_register_8(&devCtx->hw.control->CCON, 1);
  os_write_register_8(&devCtx->hw.control->CCON, 0);
  os_write_register_8(&devCtx->hw.control->CCON, 1);

  /* now send the data bits */
  for (i = 0; i < 24; i++, clockword >>= 1) {
    if (clockword & 1) {
      d = 0;
    } else {
      d = 2;
    }
		
    os_write_register_8(&devCtx->hw.control->CCON, (os_uint8) (d | 0x1U));
    os_write_register_8(&devCtx->hw.control->CCON, (os_uint8) d);
    os_write_register_8(&devCtx->hw.control->CCON, (os_uint8) (d ^ 0x2U));
    os_write_register_8(&devCtx->hw.control->CCON, (os_uint8) (d ^ 0x2U) | 0x1U);
  }

  os_write_register_8(&devCtx->hw.control->CCON, 3);
  os_write_register_8(&devCtx->hw.control->CCON, 2);
  os_write_register_8(&devCtx->hw.control->CCON, 3);
  os_write_register_8(&devCtx->hw.control->CCON, 0);

  return (void*) TRUE;
}

static OS_STATUS
buildClockWordICD2061(
  os_uint32*     clockword,
  int            index,
  unsigned long  fref,
  unsigned long  freq,
  unsigned long* actual)
{
  unsigned long W;      /* programming word for ICD2061 */
  unsigned long fvco;
  unsigned long PQERR;  /* Error so far */
  unsigned long temp;
  unsigned long act = 0;
  unsigned long I;      /* VCO range */
  unsigned long D;      /* Post-VCO divider */
  unsigned long P;      /* VCO mult */
  unsigned long Q;      /* VCO divide */
  unsigned long BP = 0;
  unsigned long	BQ = 0; /* Best so far */
  
  if (freq < 400000 || freq > 100000000) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  /* First find a post-VCO divisor */
  for (D = 0; D < 8; D++) {
    fvco = freq * (1 << D);
    if (fvco > 49990000) {
      break;
    }
  }

  for (I = 0; I < 16; I++) {
    if ((fvco > mclkRange[I][0]) && (fvco < mclkRange[I][1])) {
      break;
    }
  }

  /* Brute force P and Q guessing */
  PQERR = 1000000;
  for (Q = 3; Q < 129; Q++) {
    for (P = 4; P < 131; P++) {
      unsigned long err;

      /*
      ** The associativity of multiplication and divide operators is
      ** left to right. IF THIS FORMULA IS MODIFIED, ENSURE THAT IT
      ** CANNOT OVERFLOW AN (unsigned long).
      */
      temp = fref * P / Q / (1 << D) * 2;
      if (temp > freq) {
	err = temp - freq;
      } else {
	err = freq - temp;
      }
      if (err < PQERR) {
	BQ = Q;
	BP = P;
	PQERR = err;
	act = temp;
      }
    }
  }
  
  /* Assemble the completed word apart from the register selection */
  W = (I << 17) | ((BP - 3) << 10) | (D << 7) | (BQ - 2);
	
  /*
  ** Add the ICD2061 register index
  **
  ** index = 0 means MCLK register
  ** index = 1 means VCLK1 register
  */
  switch (index) {
  case 0:
    W |= (3 << 21);  /* ICD2061's MCLK */
    break;
  case 1:	
    W |= (0 << 21);  /* ICD2061's VCLK1 */
    break;
  default:
    break;
  }

  *clockword = W;
  *actual = act;

  return OS_STATUS_SUCCESS;
}

static OS_STATUS
programICD2061(
  ADMXRC*   devCtx,
  os_uint32 clockword)
{
  OS_STATUS status;
  int       try;

  for (try = 0; try < ADMXRC_PROGRAM_CLOCK_RETRY_LIMIT; try++) {
    os_dbgprint(3, ("programICD2061: about to call os_invoke_atomic\n"));
    devCtx->setclock.clockword = clockword;
    os_invoke_atomic(devCtx->irqObj, programICD2061Atomic, (void*) devCtx);
    
    if (os_read_register_8(&devCtx->hw.control->CCON) & 0x10) {
      os_dbgprint(0, ("programICD2061: failed\n"));
    } else {
      os_dbgprint(3, ("programICD2061: successful\n"));
      break;
    }
  }

  if (try == ADMXRC_PROGRAM_CLOCK_RETRY_LIMIT) {
    os_dbgprint(0, ("programICD2061: retry limit (%ld) exceeded\n",
		    (unsigned long) ADMXRC_PROGRAM_CLOCK_RETRY_LIMIT));
    status = OS_MAKE_STATUS(ADMXRC_FAILED);
  } else {
    status = OS_STATUS_SUCCESS;
  }

  return status;
}

OS_STATUS
ADMXRC_SetClock(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_SETCLOCK* iptr;
  os_uint32               clockword, reg;

  iptr = (ADMXRC_IOCTLS_SETCLOCK*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if ((devCtx->info.card.BoardType != ADMXRC2_BOARD_ADMXRC) &&
      (devCtx->info.card.BoardType != ADMXRC2_BOARD_ADMXRC_P)) {
    /* Allow this IOCTL to succeed only on ADM-XRC/ADM-XRC-P */
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  }
  
  clockword = (os_uint32) iptr->in.clockword;
  if (!(clockword & 0x800000U)) {
    /*
    ** Writing to one of the VCLKx or MCLK VCO registers -
    ** reject attempts to change any register other than
    ** MCLK or VCLK1
    */
    reg = (clockword >> 21) & 0x7U;
    if (reg != ADMXRC_MCLK && reg != ADMXRC_VCLK1) {
      return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    }
    
    if (devCtx->info.modifications & ADMXRC_MOD_CLOCKSWAP) {
      /*
      ** The VCLK1 and MCLK swap modification is present
      ** on this board, so we have to redirect writes to
      ** MCLK and VCLK1
      */
      switch (reg)
	{
	case ADMXRC_VCLK1:
	  reg = ADMXRC_MCLK;
	  break;
	case ADMXRC_MCLK:
	  reg = ADMXRC_VCLK1;
	  break;
	}
      clockword = (clockword & 0x1fffffU) | (reg << 21);
    }
  } else {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  os_dbgprint(3, ("ADMXRC_SetClock: clockword = 0x%lx\n", (unsigned long) clockword));

  return programICD2061(devCtx, clockword);
}

OS_STATUS
ADMXRC2_SetClock(
  ADMXRC*        devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC2_IOCTLS_SETCLOCK* iptr;
  ADMXRC2_IOCTLS_SETCLOCK* optr;
  OS_STATUS                status;
  int                      index;
  unsigned long	           freq, actual;
  os_uint32                clockword;

  iptr = (ADMXRC2_IOCTLS_SETCLOCK*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC2_IOCTLS_SETCLOCK*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  index = iptr->in.clock_index;
  freq = iptr->in.freq;

  switch (devCtx->info.card.BoardType) {
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC:
    {
      unsigned long fref;
			
      if (index > 1) {
	return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
      }

      /*
      ** index = 0 means LCLK pin on FPGA
      ** index = 1 means MCLK pin on FPGA
      */
      if (!(devCtx->info.modifications & ADMXRC_MOD_CLOCKSWAP)) {
	/*
	** The VCLK1 and MCLK swap modification is not present
	** on this board.
	*/
	index = index ^ 1;
      }

      switch (devCtx->info.clockType) {
      case ADMXRC_CLOCKTYPE_16:
	fref = 16666667;
	break;
      case ADMXRC_CLOCKTYPE_14:
	fref = 14318181;
	break;
      default:
	os_dbgprint(3, ("ADMXRC2_SetClock: invalid reference clock type (%ld)\n",
			(unsigned long) devCtx->info.clockType));
	return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
      }
      
      status = buildClockWordICD2061(&clockword, index, fref, freq, &actual);
      if (!OS_SUCCESS(status)) {
	break;
      }

      os_dbgprint(3, ("ADMXRC2_SetClock: clockword = 0x%lx actual = %ld\n",
		      (unsigned long) clockword, (unsigned long) actual));

      status = programICD2061(devCtx, clockword);
      break;
    }
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
    {
      unsigned long fref;

      if (index > 1) {
	return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
      }

      switch (devCtx->info.clockType) {
      case ADMXRC_CLOCKTYPE_16:
	fref = 16666667;
	break;
      case ADMXRC_CLOCKTYPE_14:
	fref = 14318181;
	break;
      default:
	os_dbgprint(3, ("ADMXRC2_SetClock: invalid reference clock type (%ld)\n",
			(unsigned long) devCtx->info.clockType));
	return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
      }
      
      status = buildClockWordICD2061(&clockword, index, fref, freq, &actual);
      if (!OS_SUCCESS(status)) {
	break;
      }

      os_dbgprint(3, ("ADMXRC2_SetClock: clockword = 0x%lx actual = %ld\n",
		      (unsigned long) clockword, (unsigned long) actual));

      status = programICD2061(devCtx, clockword);
      break;
    }
  default:
    os_dbgprint(3, ("ADMXRC2_SetClock: invalid board type (%ld)\n",
		    (unsigned long) devCtx->info.card.BoardType));
    return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
  }

  if (OS_SUCCESS(status)) {
    optr->out.actual = actual;
    req->status.length = sizeof(optr->out);
  }

  return status;
}

